import pickle
from pathlib import Path

import numpy
import pytest

from meshRW import msh, msh2

# load current path
CurrentPath = Path(__file__).parent
DataPath = CurrentPath / Path('test_data')
# data file for testing
datafile = DataPath / Path('debug.h5')
# artifacts directory
ArtifactsPath = CurrentPath / Path('artifacts')
ArtifactsPath.mkdir(exist_ok=True)


def test_MSHwriterTemporal():
    # open data
    hf = open(datafile, 'rb')
    #
    data = pickle.load(hf)
    hf.close()
    # extract nodes list
    nodes = data['n']
    # extract elements list and data
    elemsData = data['e']
    # generate data on nodes
    dataNodes = numpy.random.rand(nodes.shape[0], nodes.shape[1])
    # generate data on elements
    dataElem = numpy.random.rand(elemsData['TET4'].shape[0] + elemsData['PRI6'].shape[0], 1)
    # generate steps
    dataElemStep = [numpy.random.rand(elemsData['TET4'].shape[0] + elemsData['PRI6'].shape[0], 3) for i in range(5)]
    # write msh file
    outputfile = ArtifactsPath / Path('build-temp.msh')
    msh.mshWriter(
        filename=outputfile,
        nodes=nodes,
        elements=[
            {
                'connectivity': elemsData[list(elemsData.keys())[0]],
                'type': list(elemsData.keys())[0],
                'physgrp': [5, 5],
            },
            {
                'connectivity': elemsData[list(elemsData.keys())[1]],
                'type': list(elemsData.keys())[1],
                'physgrp': [6, 6],
            },
        ],
        fields=[
            {
                'data': dataNodes,
                'type': 'nodal',
                'dim': 3,
                'name': 'nodal3',
            },  # ,'steps':list of steps,'nbsteps':number of steps],
            {'data': dataElem, 'type': 'elemental', 'dim': 1, 'name': 'name_2'},
            {'data': dataElemStep, 'type': 'elemental', 'dim': 3, 'name': 'alongsteps', 'nbsteps': 5},
        ],  # ,'steps':list of steps,'nbsteps':number of steps]
    )
    assert outputfile.exists()


def test_MSHwriterStatic():
    # open data
    hf = open(datafile, 'rb')
    #
    data = pickle.load(hf)
    hf.close()
    # extract nodes list
    nodes = data['n']
    # extract elements list and data
    elemsData = data['e']
    # generate data on nodes
    dataNodes = numpy.random.rand(nodes.shape[0], nodes.shape[1])
    # generate data on elements
    dataElem = numpy.random.rand(elemsData['TET4'].shape[0] + elemsData['PRI6'].shape[0], 1)
    # write msh file
    outputfile = ArtifactsPath / Path('build.msh')
    msh.mshWriter(
        filename=outputfile,
        nodes=nodes,
        elements=[
            {
                'connectivity': elemsData[list(elemsData.keys())[0]],
                'type': list(elemsData.keys())[0],
                'physgrp': [5, 5],
            },
            {
                'connectivity': elemsData[list(elemsData.keys())[1]],
                'type': list(elemsData.keys())[1],
                'physgrp': [6, 6],
            },
        ],
        fields=[
            {
                'data': dataNodes,
                'type': 'nodal',
                'dim': 3,
                'name': 'nodal3',
            },  # ,'steps':list of steps,'nbsteps':number of steps],
            {'data': dataElem, 'type': 'elemental', 'dim': 1, 'name': 'name_2'},
        ],
    )
    assert outputfile.exists()


@pytest.mark.parametrize('append', [True, False])
@pytest.mark.parametrize('version', [2.2, 4])
def test_MSH2writerTemporal(append, version):
    # open data
    hf = open(datafile, 'rb')
    data = pickle.load(hf)
    hf.close()
    # extract nodes list
    nodes = data['n']
    # extract elements list and data
    elemsData = data['e']
    # generate data on nodes
    dataNodes = numpy.random.rand(nodes.shape[0], nodes.shape[1])
    # generate data on elements
    dataElem = numpy.random.rand(elemsData['TET4'].shape[0] + elemsData['PRI6'].shape[0], 2)
    # generate steps
    dataElemStep = [numpy.random.rand(elemsData['TET4'].shape[0] + elemsData['PRI6'].shape[0], 3) for i in range(5)]
    # write msh file
    outputfile = ArtifactsPath / Path(f'build{version}-app{append}-temp.msh')
    msh2.mshWriter(
        filename=outputfile,
        nodes=nodes,
        elements=[
            {
                'connectivity': elemsData[list(elemsData.keys())[0]],
                'type': list(elemsData.keys())[0],
                'physgrp': [5, 5],
            },
            {
                'connectivity': elemsData[list(elemsData.keys())[1]],
                'type': list(elemsData.keys())[1],
                'physgrp': [6, 6],
            },
        ],
        fields=[
            {
                'data': dataNodes,
                'type': 'nodal',
                'dim': 3,
                'name': 'nodal3',
            },  # ,'steps':list of steps,'nbsteps':number of steps],
            {'data': dataElem, 'type': 'elemental', 'dim': 2, 'name': 'name_2'},
            {'data': dataElemStep, 'type': 'elemental', 'dim': 3, 'name': 'alongsteps', 'nbsteps': 5},
        ],  # ,'steps':list of steps,'nbsteps':number of steps]
        append=append,
        opts={'version': version},
    )

    assert outputfile.exists()


@pytest.mark.parametrize('append', [True, False])
@pytest.mark.parametrize('version', [2.2, 4])
def test_MSH2writerStatic(append, version):
    # open data
    hf = open(datafile, 'rb')
    data = pickle.load(hf)
    hf.close()
    # extract nodes list
    nodes = data['n']
    # extract elements list and data
    elemsData = data['e']
    # generate data on nodes
    dataNodes = numpy.random.rand(nodes.shape[0], nodes.shape[1])
    # generate data on elements
    dataElem = numpy.random.rand(elemsData['TET4'].shape[0] + elemsData['PRI6'].shape[0], 2)
    # write msh file
    outputfile = ArtifactsPath / Path(f'build{version}-app{append}.msh')
    msh2.mshWriter(
        filename=outputfile,
        nodes=nodes,
        elements=[
            {
                'connectivity': elemsData[list(elemsData.keys())[0]],
                'type': list(elemsData.keys())[0],
                'physgrp': [5, 5],
            },
            {
                'connectivity': elemsData[list(elemsData.keys())[1]],
                'type': list(elemsData.keys())[1],
                'physgrp': [6, 6],
            },
        ],
        fields=[
            {
                'data': dataNodes,
                'type': 'nodal',
                'dim': 3,
                'name': 'nodal3',
            },  # ,'steps':list of steps,'nbsteps':number of steps],
            {'data': dataElem, 'type': 'elemental', 'dim': 2, 'name': 'name_2'},
        ],
        append=append,
        opts={'version': version},
    )

    assert outputfile.exists()


def test_MSHreader3D():
    inputfile = DataPath / Path('mesh3Dref.msh')
    # open file and read it
    mesh = msh.mshReader(filename=inputfile)
    assert mesh.getNodes().shape == (21050, 3)
    assert mesh.getTags() == [5, 6]
    assert mesh.getElements().get('TET4').shape == (96925, 4)
    assert mesh.getElements().get('PRI6').shape == (1667, 6)
    assert mesh.getElements(type='TET4').shape == (96925, 4)
    assert mesh.getElements(type='PRI6').shape == (1667, 6)
    assert mesh.getElements(tag=5).get('TET4').shape == (96925, 4)
    assert mesh.getElements(tag=6).get('PRI6').shape == (1667, 6)
    assert mesh.getElements(tag=6).get('TET4') is None
    assert mesh.getElements(tag=5).get('PRI6') is None
    assert len(mesh.getElements(tag=6, type='TET4')) == 0
    assert len(mesh.getElements(tag=5, type='PRI6')) == 0


def test_MSHreader2D():
    inputfile = DataPath / Path('mesh2Dref.msh')
    # open file and read it
    mesh = msh.mshReader(filename=inputfile)
    assert mesh.getNodes().shape == (7480, 3)
    assert mesh.getTags() == [2, 4, 1, 15, 5, 27]
    assert mesh.getElements().get('LIN2').shape == (52, 2)
    assert mesh.getElements().get('TRI3').shape == (14614, 3)
    assert mesh.getElements(type='LIN2').shape == (52, 2)
    assert mesh.getElements(type='TRI3').shape == (14614, 3)
    assert mesh.getElements(tag=2, type='LIN2').shape == (52, 2)
    assert mesh.getElements(tag=4, type='LIN2').shape == (52, 2)
    assert mesh.getElements(tag=1, type='TRI3').shape == (14614, 3)
    assert mesh.getElements(tag=15, type='TRI3').shape == (14156, 3)
    assert mesh.getElements(tag=5, type='TRI3').shape == (458, 3)
    assert mesh.getElements(tag=27, type='TRI3').shape == (458, 3)
    assert len(mesh.getElements(tag=5, type='LIN2')) == 0
    assert len(mesh.getElements(tag=27, type='LIN2')) == 0


# # if __name__ == "__main_":

# CurrentPath = os.path.dirname(__file__)
# # hf = open(os.path.abspath(os.path.join(CurrentPath,
# #                                           './test_data/debug.h5')), 'rb')

# # data = pickle.load(hf)
# # hf.close()
# # nodes = data['n']
# # elemsData = data['e']

# # dataNodes=numpy.random.rand(nodes.shape[0],nodes.shape[1])
# # dataElem=numpy.random.rand(elemsData['TET4'].shape[0]+elemsData['PRI6'].shape[0],2)
# # dataElemStep=[numpy.random.rand(elemsData['TET4'].shape[0]+elemsData['PRI6'].shape[0],3) for i in range(5)]
# # msh.mshWriter(filename=os.path.abspath(os.path.join(CurrentPath,
# #                                           './test_data/vv.msh')),
# #           nodes=nodes,
# #           elems=[{'connectivity': elemsData[list(elemsData.keys())[0]], 'type':list(elemsData.keys())[0], 'physgrp':[5, 5]},
# #                  {'connectivity': elemsData[list(elemsData.keys())[1]], 'type':list(elemsData.keys())[1], 'physgrp':[6, 6]}],
# #                  fields=[{'data':dataNodes,'type':'nodal','dim':3,'name':'nodal3'},#,'steps':list of steps,'nbsteps':number of steps],
# #                     {'data':dataElem,'type':'elemental' ,'dim':2,'name':'name_2'},
# #                     {'data':dataElemStep,'type':'elemental' ,'dim':3,'name':'alongsteps','nbsteps':5}]#,'steps':list of steps,'nbsteps':number of steps]
# #                     )


# # msh.mshReader(filename=os.path.abspath(os.path.join(CurrentPath,'./meshRW/test_data/vv.msh')))
# M = msh.mshReader(filename=os.path.abspath(os.path.join(CurrentPath,'meshRW/test_data/mesh2Dref.msh')),dim=2)
# M.getElements(dictFormat = False)
# M.getNodes()



""" "
This file includes the definition
and tools to manipulate MSH format
Documentation available here:
https://gmsh.info/doc/texinfo/gmsh.html#MSH-file-format
----
Luc Laurent - luc.laurent@lecnam.net -- 2021

"""

from loguru import logger as Logger


def loadElementDict():
    """
    dictionary from element (string) to msh element number
    """
    elementDict = {
        # 2-nodes line
        'LIN2': {'code': 1, 'nodes': 2, 'dim': 1},
        # 3-nodes second order line
        'LIN3': {'code': 8, 'nodes': 3, 'dim': 1},
        # 4-nodes third order line
        'LIN4': None,
        # 3-nodes triangle
        'TRI3': {'code': 2, 'nodes': 3, 'dim': 2},
        # 6-nodes second order triangle (3 vertices, 3 on edges)
        'TRI6': {'code': 9, 'nodes': 6, 'dim': 2},
        # 9-nodes cubic order triangle (3 vertices, 3 on edges and 3 inside)
        'TRI9': None,
        # 10-nodes higher order triangle (3 vertices, 6 on edges and 1 inside)
        'TRI10': None,
        # 12-nodes higher order triangle (3 vertices and 9 on edges)
        'TRI12': None,
        # 15-nodes higher order triangle (3 vertices, 9 on edges and 3 inside)
        'TRI15': None,
        # 4-nodes quadrangle
        'QUA4': {'code': 3, 'nodes': 4, 'dim': 2},
        # 8-nodes second order quadrangle (4 vertices and 4 on edges)
        'QUA8': {'code': 16, 'nodes': 8, 'dim': 2},
        # 9-nodes higher order quadrangle (4 vertices, 4 on edges and 1 inside)
        'QUA9': {'code': 10, 'nodes': 9, 'dim': 2},
        # 4-nodes tetrahedron
        'TET4': {'code': 4, 'nodes': 4, 'dim': 3},
        # 10-nodes second order tetrahedron (4 vertices and 6 on edges)
        'TET10': {'code': 11, 'nodes': 10, 'dim': 3},
        # 8-nodes hexahedron
        'HEX8': {'code': 5, 'nodes': 8, 'dim': 3},
        # 20-nodes second order hexahedron (8 vertices and 12 on edges)
        'HEX20': {'code': 17, 'nodes': 20, 'dim': 3},
        # 27-nodes higher order hexahedron
        # (8 vertices, 12 on edges, 6 on faces and 1 inside)
        'HEX27': {'code': 12, 'nodes': 27, 'dim': 3},
        # 6-nodes prism
        'PRI6': {'code': 6, 'nodes': 6, 'dim': 3},
        # 15-nodes second order prism (6 vertices and 9 on edges)
        'PRI15': {'code': 18, 'nodes': 15, 'dim': 3},
        # 18-nodes higher order prism (6 vertices, 9 on edges and 3 on faces)
        'PRI18': {'code': 13, 'nodes': 18, 'dim': 3},
        # 5-node pyramid
        'PYR5': {'code': 7, 'nodes': 5, 'dim': 3},
        # 13-nodes second order pyramid (5 edges and 8 on edges)
        'PYR13': {'code': 19, 'nodes': 13, 'dim': 3},
        # 14-nodes higher order pyramid (5 edges, 8 on edges and 1 inside)
        'PYR14': {'code': 14, 'nodes': 14, 'dim': 3},
        # 1-node point
        'NOD1': {'code': 15, 'nodes': 1, 'dim': 0},
    }
    return elementDict


def getMSHElemType(txtEltype):
    """
    Get the element type defined as in gmsh (refer to gmsh documentation for numbering) from text declaration
    syntax:
        getMshElemType(txtEltype)

    input:
        txtEltype: element declared using string (if number is used the function will return it)
    output:
        element type defined as gmsh number
    """
    elementDict = loadElementDict()

    # depending on the type of txtEltype
    # - if int: return txtEltype
    # - else get the number from the dictionary
    if isinstance(txtEltype, int):
        elementNum = txtEltype
    else:
        elementNum = elementDict[txtEltype.upper()].get('code', None)
    # show error if the type is not available
    if not elementNum:
        Logger.error(f'Element type {txtEltype} not implemented')
    return elementNum


def getElemTypeFromMSH(elementNum):
    """
    Get the element type from id (integer) defined in gmsh (refer to gmsh documentation for numbering)
    syntax:
        getElemTypeFromMSH(elementNum)

    input:
        elementNum: integer used in gmsh to declare element
    output:
        global name of the element
    """
    # load the dictionary
    elementDict = loadElementDict()
    globalName = None
    # get the name of the element using the integer iD along the dictionary
    for k, v in elementDict.items():
        if v:
            if v.get('code', None) == elementNum:
                globalName = k
                break
    # if the name of the element if not available show error
    if globalName is None:
        Logger.error(f'Element type not found with id {elementNum}')
    return globalName


def getNumberNodes(txtElemtype):
    """
    Get the number of nodes for a specific element type type (declare as string)
        syntax:
            getNumberNodes(txtElemtype)

       input:
            txtElemtype: element declared using string (if number is used the function wil return it)
        output:
            number of nodes for txtEltype
    """
    # load the dictionary
    elementDict = loadElementDict()
    nbNodes = 0
    # check if the type of element exists
    if txtElemtype in elementDict.keys():
        # get the number of nodes for the type of element
        if elementDict[txtElemtype]:
            nbNodes = elementDict[txtElemtype].get('nodes', None)
    else:
        # show error message if the type of element does not exist
        Logger.error(f'Element type {txtElemtype} not defined')
    return nbNodes


def getNumberNodesFromNum(elementNum):
    """
    Get the number of nodfs for a specific element type type (declare as string)
        syntax:
            getNumberNodesFromNum(elementNum)

        input:
            elementNum: integer used in gmsh to declare element
        output:
            number of nodes for txtEltype
    """
    return getNumberNodes(getElemTypeFromMSH(elementNum))


# DEFAULT VALUES
ALLOWED_EXTENSIONS = ['.msh', '.msh.bz2', '.msh.gz']

# Keywords MSH
DFLT_FILE_OPEN_CLOSE = {'open': '$MeshFormat', 'close': '$EndMeshFormat'}
DFLT_FILE_VERSION = '2.2 0 8'
DFLT_NODES_OPEN_CLOSE = {'open': '$Nodes', 'close': '$EndNodes'}
DFLT_ELEMS_OPEN_CLOSE = {'open': '$Elements', 'close': '$EndElements'}
DFLT_FIELDS_NODES_OPEN_CLOSE = {'open': '$NodeData', 'close': '$EndNodeData'}
DFLT_FIELDS_ELEMS_OPEN_CLOSE = {'open': '$ElementData', 'close': '$EndElementData'}



"""
This class is a part of the meshRW library and will write a msh file from a mesh using gmsh API
----
Luc Laurent - luc.laurent@lecnam.net -- 2024
"""

import time
from pathlib import Path
from typing import Union

import gmsh
import numpy as np
from loguru import logger as Logger

from . import dbmsh, various, writerClass


def getViewName(view_tag):
    return gmsh.option.getString(f'View[{gmsh.view.getIndex(view_tag)}].Name')


class mshWriter(writerClass.writer):
    def __init__(
        self,
        filename: Union[str, Path] = None,
        nodes: Union[list, np.ndarray] = None,
        elements: dict = None,
        fields: Union[list, np.ndarray] = None,
        append: bool = False,
        title: str = None,
        opts: dict = {'version': 2.2},
    ):
        #
        Logger.info('Create msh file using gmsh API')
        # adapt inputs
        nodes, elements, fields = writerClass.adaptInputs(nodes, elements, fields)
        # initialization
        super().__init__(filename, nodes, elements, fields, append, title, opts)
        # load specific configuration
        self.db = dbmsh
        #
        if self.title is None:
            self.title = 'Imported mesh'
        self.modelName = self.title

        # write contents
        self.writeContents(nodes, elements, fields)

    def getAppend(self):
        """
        Obtain the adapt flag from the handler (automatic adaptation if the file exists)
        """
        return self.append

    def setOptions(self, options: dict):
        """Default options"""
        self.version = options.get('version', 2.2)

    def writeContents(self, nodes, elements, fields):
        """Write contents"""
        # initialize gmsh
        gmsh.initialize()
        gmsh.option.setNumber('Mesh.MshFileVersion', self.version)
        gmsh.option.setNumber('PostProcessing.SaveMesh', 1)  # export mesh when save fields
        # create empty entities
        gmsh.model.add(self.modelName)
        self.entities = {}
        Logger.info(f'Create {len(self.listPhysGrp)} entities for physical group')
        for g in self.listPhysGrp:
            self.entities[g] = gmsh.model.addDiscreteEntity(3)
            gmsh.model.addPhysicalGroup(3, [self.entities[g]], g, name=self.nameGrp.get(g, None))
        # add global physical group
        self.globEntity = gmsh.model.addDiscreteEntity(3)
        gmsh.model.addPhysicalGroup(3, [self.globEntity], self.globPhysGrp, name='Global')

        # add nodes
        self.writeNodes(nodes)

        # add elements
        self.writeElements(elements)

        # add fields
        if fields is not None:
            self.writeFields(fields)

        # run internal gmsh function to reclassify nodes
        gmsh.model.mesh.reclassifyNodes()

        # write msh file
        self.writeFiles()
        # clean gmsh
        gmsh.finalize()

    @various.timeit('Nodes declared')
    def writeNodes(self, nodes):
        """
        write nodes coordinates
        """
        # adapt nodes
        if isinstance(nodes, list):
            nodes = np.array(nodes)
        #
        self.nbNodes = nodes.shape[0]
        Logger.debug(f'Write {self.nbNodes} nodes')
        #
        nodesNum = np.arange(1, len(nodes) + 1)
        numFgrp = self.listPhysGrp[0]
        # add nodes to first volume entity
        gmsh.model.mesh.addNodes(3, self.entities[numFgrp], nodesNum, nodes.flatten())

    @various.timeit('Elements declared')
    def writeElements(self, elements):
        """
        write elements
        input:
            elements: lists of dict of connectivity with elements type (could be reduce to only one dictionary and elements)
                    [{'connectivity':table1,'type':eltype1,'physgrp':gpr1},{connectivity':table2,'type':eltype1,'physgrp':grp2}...]
                    or
                    {'connectivity':table1,'type':eltype1,'physgrp':gpr1}

                    'connectivity': connectivity array
                    'type': type of elements (could be a string or an integer, see getGmshElemType and  gmsh documentation)
                    'physgrp' (optional): physical group (integer or array of integers to declare the physical group of each cell)
        """
        # convert to list if dict
        if type(elements) is dict:
            elemsRun = [elements]
        else:
            elemsRun = elements
        #
        Logger.info(f'Add {self.nbElems} elements')
        for m in elemsRun:
            # get connectivity data
            typeElem = m.get('type')
            connectivity = m.get('connectivity')
            physgrp = m.get('physgrp', None)
            codeElem = dbmsh.getMSHElemType(typeElem)
            #
            Logger.info(f'Set {len(connectivity)} elements of type {typeElem}')
            gmsh.model.mesh.addElementsByType(self.globEntity, codeElem, [], connectivity.flatten())
            if physgrp is not None:
                if not isinstance(physgrp, np.ndarray) and not isinstance(physgrp, list):
                    physgrp = [physgrp]
                for p in physgrp:
                    gmsh.model.mesh.addElementsByType(self.entities[p], codeElem, [], connectivity.flatten())

    @various.timeit('Fields declared')
    def writeFields(self, fields):
        """write all fields"""
        if not isinstance(fields, list):
            fields = [fields]
        Logger.info(f'Add {len(fields)} fields')
        for f in fields:
            self.writeField(f)

    def writeField(self, field):
        """write one field"""
        # load field data
        data = field.get('data')
        name = field.get('name')
        numEntities = field.get('numEntities', None)
        nbsteps = field.get('nbsteps', 1)
        steps = field.get('steps', None)
        timesteps = field.get('timesteps', None)
        dim = field.get('dim', 0)
        typeField = field.get('type')
        #
        if not name:
            name = f'{typeField}_{self.itName}'
            self.itName += 1
        if not steps and nbsteps:
            steps = np.arange(nbsteps, dtype=int)
        if not timesteps:
            timesteps = np.zeros(nbsteps)
        if nbsteps == 1 and len(data) > 1:
            data = [data]

        # add field
        if typeField == 'nodal':
            nameTypeData = 'NodeData'
            if numEntities is None:
                numEntities = np.arange(1, self.nbNodes + 1)

        elif typeField == 'elemental':
            nameTypeData = 'ElementData'
            if numEntities is None:
                numEntities = np.arange(1, self.nbElems + 1)
        else:
            raise ValueError('typeField must be nodal or elemental')
        #
        tagView = gmsh.view.add(name)
        for s, t in zip(steps, timesteps):
            dataView = data[s]
            if len(dataView.shape) == 1:
                dataView = dataView.reshape((-1, 1))
            gmsh.view.addModelData(tagView, s, self.modelName, nameTypeData, numEntities, dataView)
            # ,
            # numComponents=dim,
            # partition=0)

    @various.timeit('File(s) written')
    def writeFiles(self):
        """Advanced writing to export mesh and fields"""
        gmsh.write(self.filename.as_posix())
        if self.getAppend():
            for t in gmsh.view.getTags():
                viewname = getViewName(t)
                starttime = time.perf_counter()
                gmsh.view.write(t, self.filename.as_posix(), append=True)
                Logger.info(
                    f'Field {viewname} save in {self.filename} ({various.convert_size(self.filename.stat().st_size)})'
                )
        else:
            it = 0
            for t in gmsh.view.getTags():
                viewname = getViewName(t)
                viewname = viewname.replace(' ', '_')
                if len(viewname) > 15:
                    viewname = viewname[0:15]
                #
                newfilename = self.getFilename(suffix=f'_view-{it}_{viewname}')
                starttime = time.perf_counter()
                gmsh.view.write(t, newfilename.as_posix(), append=False)
                Logger.info(
                    f'Data save in {newfilename} ({various.convert_size(newfilename.stat().st_size)}) - Elapsed {time.perf_counter()-starttime} s'
                )

    # def dataAnalysis(self,nodes,elems,fields):
    #     """ """
    #     self.nbNodes = len(nodes)
    #     self.nbElems = 0
    #     #
    #     self.elemPerType = {}
    #     self.elemPerGrp = {}
    #     self.nameGrp = {}
    #     #
    #     if isinstance(elems,dict):
    #         elems = [elems]
    #     #
    #     itGrpE = 0
    #     for e in elems:
    #         if e.get('type') not in self.elemPerType:
    #             self.elemPerType[e.get('type')] = 0
    #         self.elemPerType[e.get('type')] += len(e.get('connectivity'))
    #         self.nbElems += len(e.get('connectivity'))
    #         name = e.get('name','grp-{}'.format(itGrpE))
    #         itGrpE += 1
    #         if e.get('physgrp') is not None:
    #             if not isinstance(e.get('physgrp'),list) or not isinstance(e.get('physgrp'),list):
    #                 physgrp = [e.get('physgrp')]
    #             else:
    #                 physgrp = e.get('physgrp')
    #             for p in np.unique(physgrp):
    #                 if p not in self.elemPerGrp:
    #                     self.elemPerGrp[p] = 0
    #                 self.elemPerGrp[p] += len(e.get('connectivity'))
    #                 #
    #                 if p not in self.nameGrp:
    #                     self.nameGrp[p] = name
    #                 else:
    #                     self.nameGrp[p] += '-' + name
    #     #
    #     self.listPhysGrp = list(self.elemPerGrp.keys())
    #     # generate global physical group
    #     numinit = 1000
    #     numit = 50
    #     current = numinit
    #     while current in self.listPhysGrp:
    #         current += numit
    #     self.globPhysGrp = current
    #     # show stats
    #     Logger.debug('Number of nodes: {}'.format(self.nbNodes))
    #     Logger.debug('Number of elements: {}'.format(self.nbElems))
    #     Logger.debug('Number of physical groups: {}'.format(len(self.listPhysGrp)))
    #     for t,e in self.elemPerType.items():
    #         Logger.debug('Number of {} elements: {}'.format(t,e))
    #     for g in self.listPhysGrp:
    #         Logger.debug('Number of elements in group {}: {}'.format(g,self.elemPerGrp.get(g,0)))
    #     Logger.debug('Global physical group: {}'.format(self.globPhysGrp))
    #     # create artificial physical group if necessary
    #     if len(self.listPhysGrp) == 0:
    #         self.listPhysGrp = [1]



"""
Various useful tools
----
Luc Laurent - luc.laurent@lecnam.net -- 2021
"""

import time

import numpy
from loguru import logger as Logger


def convert_size(size_bytes):
    """
    Convert size in bytes to human readable size
    """
    if size_bytes == 0:
        return '0B'
    size_name = ('B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB')
    i = int(numpy.floor(numpy.log(size_bytes) / numpy.log(1024)))
    p = numpy.power(1024, i)
    s = round(size_bytes / p, 2)
    return f'{s:g} {size_name[i]}'


# decorator to measure time
def timeit(txt=None):
    def decorator(func):
        def timeit_wrapper(*args, **kwargs):
            start_time = time.perf_counter()
            result = func(*args, **kwargs)
            end_time = time.perf_counter()
            total_time = end_time - start_time
            # first item in the args, ie `args[0]` is `self`
            if txt is not None:
                Logger.debug(f'{txt} - {total_time:.4f} s')
            else:
                Logger.debug(f'Execution time: {total_time:.4f} s')
            return result

        return timeit_wrapper

    return decorator



from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import Union

import numpy as np
from loguru import logger as Logger


class writer(ABC):
    def __init__(
        self,
        filename: Union[str, Path] = None,
        nodes: Union[list, np.ndarray] = None,
        elements: dict = None,
        fields: Union[list, np.ndarray] = None,
        append: bool = False,
        title: str = None,
        opts: dict = {},
    ):
        self.append = append
        self.title = self.adaptTitle(txt=title)
        self.filename = Path(filename)
        self.basename = self.filename.name
        # set options
        self.setOptions(opts)
        #
        self.db = None
        #
        self.nbNodes = 0
        self.nbElems = 0
        #
        self.listPhysGrp = []
        self.nbSteps = 0
        self.steps = []
        self.nbFields = 0
        # run data analysis
        self.dataAnalysis(nodes, elements, fields)


    @abstractmethod
    def setOptions(self, opts):
        """Set options"""

    @abstractmethod
    def getAppend(self):
        pass

    def adaptTitle(self, txt='', append=False):
        """Adapt title with additional information"""
        if append:
            txtFinal = self.title + txt
        else:
            txtFinal = txt
        if not txtFinal:
            txtFinal = datetime.today().strftime('%Y-%M-%d %H:%M:%s')
        return txtFinal

    @abstractmethod
    def writeContents(self, nodes, elements, fields=None, numStep=None):
        """Write contents"""

    def writeHeader(self):
        """Write header to the output file"""

    @abstractmethod
    def writeNodes(self, nodes):
        """write nodes"""

    @abstractmethod
    def writeElements(self, elements):
        """write elements"""

    @abstractmethod
    def writeFields(self, fields, numStep=None):
        """write fields"""

    def splitFilename(self):
        """
        Get the basename and extension (in list) of the filename
        """
        extension = ''
        filename = self.filename
        it = 0
        while it < 2:
            path = self.filename.parent
            filename = self.filename.stem
            ext = self.filename.suffix
            extension += ext
            if extension in self.db.ALLOWED_EXTENSIONS:
                it = 3
            else:
                it += 1
            if it == 2:
                self.logBadExtension()
        return path, filename, extension

    def getFilename(self, prefix=None, suffix=None, extension=None):
        """
        Add prefix and/or suffix to the filename
        """
        path, basename, ext = self.splitFilename()
        if prefix is not None:
            basename = prefix + basename
        if suffix is not None:
            basename = basename + suffix
        if extension is not None:
            ext = extension
        return path / (basename + ext)

    def logBadExtension(self):
        """ """
        Logger.error('File {}: bad extension (ALLOWED: {})'.format(self.filename, ' '.join(self.db.ALLOWED_EXTENSIONS)))

    def dataAnalysis(self, nodes, elems, fields):
        """ """
        self.nbNodes = len(nodes)
        self.nbElems = 0
        #
        self.elemPerType = {}
        self.elemPerGrp = {}
        self.nameGrp = {}
        #
        if isinstance(elems, dict):
            elems = [elems]
        #
        itGrpE = 0
        for e in elems:
            if e.get('type') not in self.elemPerType:
                self.elemPerType[e.get('type')] = 0
            self.elemPerType[e.get('type')] += len(e.get('connectivity'))
            self.nbElems += len(e.get('connectivity'))
            name = e.get('name', f'grp-{itGrpE}')
            itGrpE += 1
            if e.get('physgrp') is not None:
                if not isinstance(e.get('physgrp'), list) or not isinstance(e.get('physgrp'), list):
                    physgrp = [e.get('physgrp')]
                else:
                    physgrp = e.get('physgrp')
                for p in np.unique(physgrp):
                    if p not in self.elemPerGrp:
                        self.elemPerGrp[p] = 0
                    self.elemPerGrp[p] += len(e.get('connectivity'))
                    #
                    if p not in self.nameGrp:
                        self.nameGrp[p] = name
                    else:
                        self.nameGrp[p] += '-' + name
        #
        self.listPhysGrp = list(self.elemPerGrp.keys())
        # generate global physical group
        numinit = 1000
        numit = 50
        current = numinit
        while current in self.listPhysGrp:
            current += numit
        self.globPhysGrp = current
        # show stats
        Logger.debug(f'Number of nodes: {self.nbNodes}')
        Logger.debug(f'Number of elements: {self.nbElems}')
        Logger.debug(f'Number of physical groups: {len(self.listPhysGrp)}')
        for t, e in self.elemPerType.items():
            Logger.debug(f'Number of {t} elements: {e}')
        for g in self.listPhysGrp:
            Logger.debug(f'Number of elements in group {g}: {self.elemPerGrp.get(g,0)}')
        Logger.debug(f'Global physical group: {self.globPhysGrp}')
        # create artificial physical group if necessary
        if len(self.listPhysGrp) == 0:
            self.listPhysGrp = [1]
        ## analyse fields
        if fields is not None:
            if isinstance(fields, dict):
                fields = [fields]
            self.fieldAnalysis(fields)

    def fieldAnalysis(self, fields: list):
        """Analyse fields"""
        self.nbFields = len(fields)
        self.nbCellFields = 0
        self.nbPointFields = 0
        self.nbTemporalFields = 0
        itField = -1
        for f in fields:
            itField += 1
            if f.get('type') == 'elemental':
                self.nbCellFields += 1
            elif f.get('type') == 'nodal':
                self.nbPointFields += 1
            if f.get('nbsteps') is not None or f.get('steps') is not None:
                self.nbTemporalFields += 1
                cSteps = []
                if f.get('steps') is not None:
                    cSteps = f.get('steps')
                cNbSteps = f.get('nbsteps', len(cSteps))
                # adapt steps
                if len(self.steps) < cNbSteps:
                    cSteps = np.arange(cNbSteps, dtype=float)
                if cNbSteps == 0:
                    cNbSteps = len(self.steps)
                # check consistency of definition of steps
                if len(self.steps) > 0:
                    if not np.allclose(self.steps, cSteps):
                        name = f.get('name', f'field-{itField}')
                        Logger.error(f'Inconsistent steps in fields {name}')
                else:
                    self.steps = cSteps
                    self.nbSteps = cNbSteps

        # show stats
        Logger.debug(f'Number of fields: {self.nbFields}')
        Logger.debug(f'Number of cell fields: {self.nbCellFields}')
        Logger.debug(f'Number of point fields: {self.nbPointFields}')
        Logger.debug(f'Number of temporal fields: {self.nbTemporalFields}')


def adaptInputs(nodes, elements, fields):
    """ Adapt inputs for the writer """
    # adapt nodes
    if nodes is not None:
        if isinstance(nodes, list):
            nodes = np.array(nodes)
    else:
        Logger.error('No nodes provided')
    # adapt elements
    if elements is not None:
        if isinstance(elements, dict):
            elements = [elements]
        for e in elements:
            if e.get('connectivity') is not None:
                e['connectivity'] = np.array(e.get('connectivity'))
    else:
        Logger.error('No elements provided')
    # adapt fields
    if fields is not None:
        if isinstance(fields, dict):
            fields = [fields]
        for f in fields:
            if f.get('steps') is not None:
                f['steps'] = np.array(f.get('steps'))
            if f.get('data') is not None:
                if isinstance(f.get('data'), list):
                    f['data'] = np.array(f.get('data'))
    else:
        Logger.warning('No fields provided')

    return nodes, elements, fields


"""
This class is a part of
the SILEX library and
will write results in msh v2 file.
Documentation available here:
https://gmsh.info/doc/texinfo/gmsh.html#MSH-file-format
----
Luc Laurent - luc.laurent@lecnam.net -- 2021
"""

from pathlib import Path
from typing import Union

import numpy as np
from loguru import logger as Logger

from . import configMESH, dbmsh, fileio, various, writerClass


class mshWriter(writerClass.writer):
    """ """

    def __init__(
        self,
        filename: Union[str, Path] = None,
        nodes: Union[list, np.ndarray] = None,
        elements: Union[list, np.ndarray] = None,
        fields: Union[list, np.ndarray] = None,
        append: bool = False,
        title: str = None,
        opts: dict = {},
    ):
        """
        load class to write file

        syntax:
            mshWriter(filename,nodes,elems,fields)

        inputs:
            filename : string, name of the gmsh file (with or without msh extension),
                                may contain a directory name
            nodes    : nodes coordinates
            elements : connectivity tables (could contains many kind of elements)
                    list of connectivity dict such as [{'connectivity':table1,'type':eltype1,'physgrp':gpr1},{connectivity':table2,'type':eltype1,'physgrp':grp2}...]
                    'connectivity': connectivity array
                    'type': type of elements (could be a string or an integer, see getGmshElemType and  gmsh documentation)
                    'physgrp' (optional): physical group (integer or array of integers to declare the physical group of each cell)
            fields (optional)  : list of the fields to write, syntax:
                fields=[{'data':variable_name1,'type':'nodal' or 'elemental' 'numentities': number of concerned values (nodes, elements),'dim':number of values per node,'name':'name 1','steps':list of steps,'nbsteps':number of steps],
                        {'data':variable_name2,'type':'nodal' or 'elemental' ,'dim':number of values per node,'name':'name 2','steps':list of steps,'nbsteps':number of steps],
                        ...
                        ]
                append (optional, default: False) : append field to an existing file
                title (optional, default: None) : title of the file
        """
        Logger.info('Start writing msh file')
        # adapt inputs
        nodes, elements, fields = writerClass.adaptInputs(nodes, elements, fields)
        # initialization
        super().__init__(filename, nodes, elements, fields, append, title, opts)

        # load specific configuration
        self.db = dbmsh
        # depending on the case
        Logger.info(f'Initialize writing {self.basename}')
        if fields is not None and self.append and self.filename.exists():
            self.fhandle = fileio.fileHandler(filename=filename, right='a', safeMode=False)
        else:
            self.fhandle = fileio.fileHandler(filename=filename, right='w', safeMode=False)

        # write contents
        self.writeContents(nodes, elements, fields)

        # close file
        self.fhandle.close()
        self.fhandle = None

    def setOptions(self, options: dict):
        """Default options"""

    def writeContents(self, nodes, elements, fields=None, numStep=None):
        """Write contents"""
        if not self.getAppend():
            # write header
            self.fhandle.write('{}\n'.format(dbmsh.DFLT_FILE_OPEN_CLOSE['open']))
            self.fhandle.write(f'{dbmsh.DFLT_FILE_VERSION}\n')
            self.fhandle.write('{}\n'.format(dbmsh.DFLT_FILE_OPEN_CLOSE['close']))
            # write nodes
            self.writeNodes(nodes)
            # write elements
            self.writeElements(elements)

        # write fields
        if fields is not None:
            self.writeFields(fields)

    def getAppend(self):
        """
        Obtain the adapt flag from the handler (automatic adaptation if the file exists)
        """

        self.append = self.fhandle.append
        return self.append

    @various.timeit('Nodes written')
    def writeNodes(self, nodes):
        """
        write nodes coordinates
        """
        # adapt nodes
        if isinstance(nodes, list):
            nodes = np.array(nodes)
        #
        self.nbNodes = nodes.shape[0]
        Logger.debug(f'Write {self.nbNodes} nodes')
        #
        self.fhandle.write('{}\n'.format(dbmsh.DFLT_NODES_OPEN_CLOSE['open']))
        self.fhandle.write(f'{self.nbNodes}\n')
        #
        self.dimPb = nodes.shape[1]

        # (2d)
        if self.dimPb == 2:
            #
            formatSpec = '{:d} {:9.4g} {:9.4g} 0.0\n'
            # write
            for i in range(self.nbNodes):
                self.fhandle.write(formatSpec.format(i + 1, *nodes[i, :], 0.0))

        # (3d)
        if self.dimPb == 3:
            #
            formatSpec = '{:d} {:9.4g} {:9.4g} {:9.4g}\n'
            # write
            for i in range(self.nbNodes):
                self.fhandle.write(formatSpec.format(i + 1, *nodes[i, :]))

        self.fhandle.write('{}\n'.format(dbmsh.DFLT_NODES_OPEN_CLOSE['close']))

    @various.timeit('Elements written')
    def writeElements(self, elems):
        """
        write elements
        input:
            elems: lists of dict of connectivity with elements type (could be reduce to only one dictionary and elements)
                    [{'connectivity':table1,'type':eltype1,'physgrp':gpr1},{connectivity':table2,'type':eltype1,'physgrp':grp2}...]
                    or
                    {'connectivity':table1,'type':eltype1,'physgrp':gpr1}

                    'connectivity': connectivity array
                    'type': type of elements (could be a string or an integer, see getGmshElemType and  gmsh documentation)
                    'physgrp' (optional): physical group (integer or array of integers to declare the physical group of each cell)
        """
        # convert to list if dict
        if type(elems) is dict:
            elemsRun = [elems]
        else:
            elemsRun = elems

        # count number of elems
        self.nbElems = 0
        Logger.debug('Start preparing elements')
        for iD in elemsRun:
            dimC = iD[configMESH.DFLT_MESH].shape
            #
            iD['nbElems'] = dimC[0]  # nb of elements on the connectivity table
            iD['nbNodes'] = dimC[1]  # nb of nodes per element
            self.nbElems += dimC[0]  # total number of elements
            #
            # convert element type to MSH number
            iD['eltypeGMSH'] = dbmsh.getMSHElemType(iD[configMESH.DFLT_TYPE_ELEM])
            #
            if configMESH.DFLT_PHYS_GRP not in iD.keys():
                iD[configMESH.DFLT_PHYS_GRP] = 0
            if type(iD[configMESH.DFLT_PHYS_GRP]) is int:
                iD[configMESH.DFLT_PHYS_GRP] = [iD[configMESH.DFLT_PHYS_GRP]]
            if len(iD[configMESH.DFLT_PHYS_GRP]) == 1:
                iD[configMESH.DFLT_PHYS_GRP].append(iD[configMESH.DFLT_PHYS_GRP][0])
        Logger.debug('Done')

        # write all meshes
        Logger.debug(f'Start writing {self.nbElems} elements')
        self.fhandle.write('{}\n'.format(dbmsh.DFLT_ELEMS_OPEN_CLOSE['open']))
        self.fhandle.write(f'{self.nbElems}\n')
        itElem = 0  # iterator for elements
        for iD in elemsRun:
            # create format specifier for element
            # 1: number of element
            # 2: type of the element (see gmsh documentation)
            # 3: number of tags (minimum number=2)
            # 4: physical entity
            # 5: elementary entity
            # 6+: nodes of the elements
            formatSpec = ' '.join('{:d}' for i in range(3 + len(iD[configMESH.DFLT_PHYS_GRP]) + iD['nbNodes'])) + '\n'
            # write
            for e in range(iD['nbElems']):
                itElem += 1
                # write in file
                self.fhandle.write(
                    formatSpec.format(
                        itElem,
                        iD['eltypeGMSH'],
                        len(iD[configMESH.DFLT_PHYS_GRP]),
                        *iD[configMESH.DFLT_PHYS_GRP],
                        *iD[configMESH.DFLT_MESH][e],
                    )
                )

        self.fhandle.write('{}\n'.format(dbmsh.DFLT_ELEMS_OPEN_CLOSE['close']))

    @various.timeit('Fields written')
    def writeFields(self, fields):
        """
        write fields
        input:
            elems: lists of dict of connectivity with elements type (could be reduce to only one dictionary and elements)
                    [{'connectivity':table1,'type':eltype1,'physgrp':gpr1},{connectivity':table2,'type':eltype1,'physgrp':grp2}...]
                    or
                    {'connectivity':table1,'type':eltype1,'physgrp':gpr1}

                    'connectivity': connectivity array
                    'type': type of elements (could be a string or an integer, see getGmshElemType and  gmsh documentation)
                    'physgrp' (optional): physical group (integer or array of integers to declare the physical group of each cell)
            fields=[{'data':variable_name1,'type':'nodal' or 'elemental' ,'dim':number of values per node,'name':'name 1','steps':list of steps,'nbsteps':number of steps],
                        {'data':variable_name2,'type':'nodal' or 'elemental' ,'dim':number of values per node,'name':'name 2','steps':list of steps,'nbsteps':number of steps],
                        ...
                        ]

                    'data': array of the data or list of dictonary
                    'type': ('nodal' or 'elemental') data given at nodes or cells
                    'dim': number of data per nodes/cells
                    'name': name of the data
                    'steps' (optional): list of steps used to declare fields
                    'nbsteps' (optional): number of steps used to declare fields
                    if no 'steps' or 'nbsteps' are declared the field is assumed to be not defined along steps
                    #
                    'data' could be defined as
                         - list of a arrays with all nodal or elemental values along steps
                         - a dictionary {'array':ar,'connectivityId':int} in the case of elemental
                            'connectivityId': the data are given associated to a certain list of cells (other is defined as 0)
        """
        # convert to list if dict
        if fields is dict:
            fieldsRun = [fields]
        else:
            fieldsRun = fields

        # along data
        Logger.debug('Start writing fields')
        for iF in fieldsRun:
            nameField = iF[configMESH.DFLT_FIELD_NAME]
            # number of data per nodes/cells
            nbPerEntity = iF[configMESH.DFLT_FIELD_DIM]
            if configMESH.DFLT_FIELD_STEPS in iF.keys():
                listSteps = iF[configMESH.DFLT_FIELD_STEPS]
                nbSteps = len(listSteps)
            elif configMESH.DFLT_FIELD_NBSTEPS in iF.keys():
                nbSteps = iF[configMESH.DFLT_FIELD_NBSTEPS]
                listSteps = range(nbSteps)
            else:
                nbSteps = 1
                listSteps = [0.0]
            Logger.debug(f'Field: {nameField}, number of steps: {nbSteps}, dimension per node/cell: {nbPerEntity}')
            # reformat values as list of arrays
            if len(iF[configMESH.DFLT_FIELD_DATA]) > 1 and nbSteps == 1:
                values = [iF[configMESH.DFLT_FIELD_DATA]]
            else:
                values = iF[configMESH.DFLT_FIELD_DATA]
            # format specifier to write fields
            formatSpec = '{:d} ' + ' '.join('{:9.4f}' for i in range(nbPerEntity)) + '\n'
            # along steps
            for iS in range(nbSteps):
                if nbSteps > 1:
                    Logger.debug(f'Step number: {iS+1}/{nbSteps}')
                if iF[configMESH.DFLT_FIELD_TYPE] == configMESH.DFLT_FIELD_TYPE_NODAL:
                    typeData = dbmsh.DFLT_FIELDS_NODES_OPEN_CLOSE
                elif iF[configMESH.DFLT_FIELD_TYPE] == configMESH.DFLT_FIELD_TYPE_ELEMENT:
                    typeData = dbmsh.DFLT_FIELDS_ELEMS_OPEN_CLOSE
                self.fhandle.write('{}\n'.format(typeData['open']))
                self.fhandle.write('1\n')  # one string tag
                # the name of the view
                self.fhandle.write(f'"{nameField}"\n')
                self.fhandle.write('1\n')  # one real tag
                self.fhandle.write(f'{listSteps[iS]:9.4f}\n')  # the time value
                self.fhandle.write('3\n')  # three integer tags
                self.fhandle.write(f'{iS:d}\n')  # time step value
                # number of components per nodes
                self.fhandle.write(f'{nbPerEntity:d}\n')
                # number of nodal values
                self.fhandle.write(f'{values[iS].shape[0]:d}\n')
                #
                for i in range(values[iS].shape[0]):
                    self.fhandle.write(formatSpec.format(i + 1, *values[iS][i, :]))

                self.fhandle.write('{}\n'.format(typeData['close']))


class mshReader:
    def __init__(self, filename=None, type='mshv2', dim=3):
        self.initcontent()
        Logger.debug(f'Open file {filename}')
        # open file and get handle
        self.objFile = fileio.fileHandler(filename=filename, right='r', safeMode=False)
        self.fhandle = self.objFile.getHandler()
        # read file line by line
        for line in self.fhandle:
            if not self.readData:
                self.readData = catchTag(line)
            elif self.readData == 'nodes':
                # read nodes
                self.readNodes(dim, line)
            elif self.readData == 'elems':
                # read elements
                self.readElements(line)
        # finalize data
        self._finalizeElems()

        # close file
        self.objFile.close()

    def initcontent(self):
        self.nodes = None  # array of nodes coordinates
        self.dim = None  # dimension of the mesh (2/3)
        self.nbNodes = None  # number of nodes
        self.elems = {}  # dictionary of elements (keys are name of the element)
        self.tagsList = {}  # list of tags and associated elements
        self.fhandle = None
        self.objFile = None
        self.readData = None
        self.curIt = 0

    def __del__(self):
        self.clean()

    def clean(self):
        """Clean the object"""
        self.initcontent()

    def readNodes(self, dim=None, lineStr=None):
        """
        Read nodes in msh file from line
            imputs:
                - dim (optional): dimension of the nodes (2D/3D)
                - lineStr: content of the current line
        """
        contentLine = lineStr.split()
        # first read: access to the number of nodes
        if self.curIt == 0:
            # read number of nodes
            self.nbNodes = int(contentLine[0])
            self.curIt += 1
            self.dim = dim
            Logger.debug(f'Start read {self.nbNodes} nodes')
        else:
            if self.curIt == 1:
                if not dim:
                    # extract dimension
                    self.dim = len(contentLine) - 1
                # create array to store nodes coordinates
                self.nodes = np.zeros((self.nbNodes, self.dim))
            # store nodes
            self.nodes[self.curIt - 1, :] = np.array(contentLine[1 : self.dim + 1], dtype=float)
            self.curIt += 1
            # stop read nodes
            if self.curIt - 1 == self.nbNodes:
                self.readData = None
                self.curIt = 0
                Logger.debug(f'Nodes read: {self.nbNodes}, dimension: {self.dim}')

    def readElements(self, lineStr=None):
        """
        Read elements
        imput:
                - lineStr: content of the current line
        """

        contentLine = lineStr.split()
        # first read: access to the number of elements
        if self.curIt == 0:
            # read number of elements
            self.nbElems = int(contentLine[0])
            self.curIt += 1
            Logger.debug(f'Start read {self.nbElems} elements')
        else:
            # store elements connectivity
            self._readElementsLine(contentLine)
            self.curIt += 1
            # stop read elements
            if self.curIt - 1 == self.nbElems:
                #  finalize data
                self._finalizeElems()
                # reset
                self.readData = None
                self.curIt = 0
                Logger.debug(f'Elements read: {self.nbElems}')
                Logger.debug('Type of elements')
                for elemType in self.elems.keys():
                    Logger.debug(f' > {self.elems[elemType].shape[0]} {elemType}')
                Logger.debug('Tags')
                for tagName in self.tagsList.keys():
                    Logger.debug(f'Tag: {tagName}')
                    for elemType in self.tagsList[tagName].keys():
                        Logger.debug(f' > {len(self.tagsList[tagName][elemType])} {elemType}')

    def _finalizeElems(self):
        """
        finalize data
        """
        for iT in self.elems:
            self.elems[iT] = np.array(self.elems[iT])

    def _readElementsLine(self, arraystr):
        """
        Read a line for elements
        input:
            - arraystr: content of a line for element (array of string)
        """
        # convert to int
        arrayint = np.array(arraystr, dtype=int)
        # get element id
        elementId = arrayint[1]
        elemType = dbmsh.getElemTypeFromMSH(elementId)
        # get number of tags
        nbTags = arrayint[2]
        tags = arrayint[3 : 3 + nbTags]
        # element nodes
        elementNodes = arrayint[3 + nbTags :]
        # check if element can be stored
        if elemType not in self.elems.keys():
            self.elems[elemType] = list()
        # store it
        self.elems[elemType].append(elementNodes)
        # get the item of the element
        iX = len(self.elems[elemType]) - 1
        # create the list for each tag
        for iT in tags:
            iTs = str(iT)
            # check if the tag already exists
            if iTs not in self.tagsList.keys():
                self.tagsList[iTs] = dict()
            # check if the element type has been already created
            if elemType not in self.tagsList[iTs].keys():
                self.tagsList[iTs][elemType] = list()
            # store the elements
            self.tagsList[iTs][elemType].append(iX)

    def getNodes(self):
        """
        Return the array of nodes coordinates
        """
        return self.nodes

    def getElements(self, type=None, tag=None, dictFormat=True):
        """
        Return elements list
        Inputs:
            - type (optional): choice the type of elements (using gmsh element id or general name of elements)
            - tag (option): choice of the tag (to export a specific part of the mesh)
        """
        elemsTag = dict()
        # filter by tag
        if tag:
            if str(tag) in self.tagsList.keys():
                elemsTag = self.tagsList[str(tag)]
        else:
            # copy all meshes associated to all tags
            # along tags
            for iT, vT in self.tagsList.items():
                # along meshes in tag
                for iM in vT.keys():
                    # check if element type already exists
                    if iM not in elemsTag.keys():
                        elemsTag[iM] = list()
                    elemsTag[iM].extend(vT[iM])
        # filter by type
        if type:
            elemsExport = list()
            elemsExportUnique = list()
            if type in elemsTag.keys():
                elemsExport = self.elems[type][elemsTag[type], :]
                elemsExportUnique = np.unique(elemsExport, axis=0)
        else:
            elemsExport = dict()
            elemsExportUnique = dict()
            for iT in elemsTag.keys():
                elemsExport[iT] = self.elems[iT][elemsTag[iT], :]
                elemsExportUnique[iT] = np.unique(elemsExport[iT], axis=0)

        # specific export
        if not dictFormat and not type:
            if len(elemsExport) > 1:
                Logger.warning('Elements exported without the dictionary format: some data are not exported')
            idElems = list(elemsExport.keys())[0]
            elemsExport = elemsExport[idElems]
            elemsExportUnique = elemsExportUnique[idElems]

        return elemsExportUnique

    def getTags(self):
        """
        Return the list of tags (integer)
        """
        listTags = list()
        listExport = list()
        if self.tagsList:
            listTags = self.tagsList.keys()
            # convert to integer
            for iL in listTags:
                listExport.append(int(iL))
        return listExport

    def getTypes(self):
        """
        Return the list of tags (integer)
        """
        listTypes = list()
        if self.elems:
            listTypes = list(self.elems.keys())
        return listTypes


def catchTag(content=None):
    """
    Try to find a specific tag in content
    """
    tagStartNodes = dbmsh.DFLT_NODES_OPEN_CLOSE['open']
    tagStartElems = dbmsh.DFLT_ELEMS_OPEN_CLOSE['open']
    typeTag = None
    if tagStartNodes == content.strip():
        typeTag = 'nodes'
    if tagStartElems == content.strip():
        typeTag = 'elems'
    return typeTag


def checkContentLine(content=None, pattern=None, item=0):
    """
    check if the pattern could be found exactly in content (list)
    """
    status = True
    if not content:
        status = False
    if not pattern:
        status = False
    if len(content) == 0:
        status = False

    if status:
        if content[0] != pattern:
            status = False
    return status



"""
This class is a part of the meshRW library and will write a vtk file from a mesh using libvtk
----
Luc Laurent - luc.laurent@lecnam.net -- 2024
"""

from pathlib import Path
from typing import Union

import numpy as np
import vtk
import vtkmodules.util.numpy_support as ns
from loguru import logger as Logger
from lxml import etree

from . import configMESH, dbvtk, various, writerClass


class vtkWriter(writerClass.writer):
    def __init__(
        self,
        filename: Union[str, Path] = None,
        nodes: Union[list, np.ndarray] = None,
        elements: dict = None,
        fields: Union[list, np.ndarray] = None,
        append: bool = False,
        title: str = None,
        opts: dict = {'binary': False, 'ascii': True},
    ):
        #
        Logger.info('Start writing vtk/vtu file using libvtk')
        # adapt inputs
        nodes, elements, fields = writerClass.adaptInputs(nodes, elements, fields)
        # prepare new fields (from physical groups for instance)
        newFields = self.createNewFields(elements)
        if newFields:
            if not fields:
                fields = list()
            fields.extend(newFields)
        # initialization
        super().__init__(filename, nodes, elements, fields, append, title, opts)
        # vtk data
        self.ugrid = None
        self.writer = None
        # load specific configuration
        self.db = dbvtk
        # write contents depending on the number of steps
        self.writeContentsSteps(nodes, elements, fields)

    def getAppend(self):
        """
        Obtain the append option
        """
        return self.append

    def setOptions(self, options: dict):
        """Default options"""
        self.binary = options.get('binary', False)
        self.ascii = options.get('ascii', False)

    def writeContentsSteps(self, nodes, elements, fields=None, numStep=None):
        """Write content along steps"""
        # create dictionary for preparing pvd file writing
        if self.nbSteps > 0:
            dataPVD = dict()
        # initialize data
        # create UnstructuredGrid
        self.ugrid = vtk.vtkUnstructuredGrid()
        # add points
        self.writeNodes(nodes)
        # elements
        self.writeElements(elements)
        # write along steps
        if self.nbSteps > 0:
            for itS in range(self.nbSteps):
                # add fieds
                self.writeContents(fields, numStep=itS)
                # adapt the filename
                filename = self.getFilename(suffix='.' + str(itS).zfill(len(str(self.nbSteps))))
                # write file
                self.write(self.ugrid, filename)
                # update PVD dict
                dataPVD[self.steps[itS]] = filename.name

                # # adapt title
                # self.title = self.adaptTitle(txt=f' step num {itS:d}', append=True)
            # write pvd file
            self.writePVD(dataPVD)
        else:
            # add fieds
            self.writeContents(fields)
            # write file
            filename = self.getFilename()
            self.write(self.ugrid, filename)

    def writePVD(self, dataPVD):
        """Write pvd file"""
        filename = self.getFilename(extension='.pvd')
        # create root element
        root = etree.Element('VTKFile', type='Collection', version='0.1')

        # Create collection elements
        collection = etree.SubElement(root, 'Collection')

        # Loop on timesteps and files for dataset
        for timestep, file in dataPVD.items():
            dataset = etree.Element('DataSet', timestep=str(timestep), part='0', file=file)
            collection.append(dataset)

        # convert xml tree to string
        xml_str = etree.tostring(root, pretty_print=True, xml_declaration=True, encoding='UTF-8')

        # write in file
        with open(filename, 'wb') as f:
            f.write(xml_str)

        Logger.debug(f"PVD file '{filename}' successfully written")

    @various.timeit('Fields declared')
    def writeContents(self, fields, numStep=None):
        """
        Add fields depending on version
        """
        self.writeFields(fields, numStep=numStep)

    def writeFields(self, fields, numStep=None):
        """Write fields"""
        if fields is not None:
            if not isinstance(fields, list):
                fields = [fields]
            Logger.info(f'Add {len(fields)} fields')
            for f in fields:
                data, typedata = self.setField(f, numStep=numStep)
                if typedata == 'nodal':
                    self.ugrid.GetPointData().AddArray(data)
                elif typedata == 'elemental':
                    self.ugrid.GetCellData().AddArray(data)

    @various.timeit('Nodes declared')
    def writeNodes(self, nodes):
        """
        Add nodes depending on version
        """
        points = vtk.vtkPoints()
        for i in range(len(nodes)):
            points.InsertNextPoint(nodes[i, :])
        self.ugrid.SetPoints(points)

    @various.timeit('Elements declared')
    def writeElements(self, elements):
        """
        Add elements depending on version
        """
        for m in elements:
            # get connectivity data
            typeElem = m.get('type')
            connectivity = m.get('connectivity')
            physgrp = m.get('physgrp', None)
            # load element's vtk class
            cell, nbnodes = dbvtk.getVTKObj(typeElem)
            Logger.debug(f'Set {len(connectivity)} elements of type {typeElem}')
            #
            for t in connectivity:
                for i in range(nbnodes):
                    cell.GetPointIds().SetId(i, t[i])
                self.ugrid.InsertNextCell(cell.GetCellType(), cell.GetPointIds())

    def createNewFields(self, elems):
        """
        Prepare new fields from elems data (for instance physical group)
        """
        # check if physgroup exists
        physGrp = False
        newFields = None
        for itE in elems:
            if configMESH.DFLT_PHYS_GRP in itE.keys():
                physGrp = True
                break
        if physGrp:
            newFields = list()
            data = list()
            for itE in elems:
                nbElems = itE[configMESH.DFLT_MESH].shape[0]
                if configMESH.DFLT_PHYS_GRP in itE.keys():
                    dataPhys = np.array(itE[configMESH.DFLT_PHYS_GRP], dtype=int)
                    if len(dataPhys) == nbElems:
                        data = np.append(data, dataPhys)
                    else:
                        data = np.append(data, dataPhys[0] * np.ones(nbElems))
                else:
                    data = np.append(data, -np.ones(nbElems))
            Logger.debug('Create new field for physical group')
            newFields.extend([{'data': data, 'type': 'elemental', 'dim': 1, 'name': configMESH.DFLT_PHYS_GRP}])

        return newFields

    def setField(self, field, numStep=None):
        """ """
        # load field data
        data = field.get('data')
        name = field.get('name')
        numEntities = field.get('numEntities', None)
        nbsteps = field.get('nbsteps', 1)
        steps = field.get('steps', None)
        dim = field.get('dim', 0)
        typeField = field.get('type')
        # for time dependent data
        if numStep is not None:
            if nbsteps > 1 or steps is not None:
                data = data[numStep]
        # initialize VTK's array
        dataVtk = ns.numpy_to_vtk(data)
        # dataVtk = vtk.vtkDoubleArray()
        dataVtk.SetName(name)
        # if len(data.shape) == 1:
        #     dim = 1
        # else:
        #     dim = data.shape[1]
        # for _,c in enumerate(data):
        #     if dim == 1:
        #         dataVtk.InsertNextValue(c)
        #     elif dim == 2:
        #         dataVtk.InsertNextTuple2(*c)
        #     elif dim == 3:
        #         dataVtk.InsertNextTuple3(*c)
        #     elif dim == 4:
        #         dataVtk.InsertNextTuple4(*c)
        #     elif dim == 6:
        #         dataVtk.InsertNextTuple6(*c)
        #     elif dim == 9:
        #         dataVtk.InsertNextTuple9(*c)
        # #
        return dataVtk, typeField

    def write(self, ugrid=None, filename=None):
        """
        Write Paraview's files along time steps
        """
        # initialization
        if self.writer is None:
            self.writer = vtk.vtkXMLUnstructuredGridWriter()
            self.writer.SetInputDataObject(self.ugrid)
            if self.binary:
                self.writer.SetFileType(vtk.VTK_BINARY)
            if self.ascii:
                self.writer.SetDataModeToAscii()
        self.writer.SetFileName(filename)
        self.writer.Update()
        # self.writer.SetDebug(True)
        # self.writer.SetWriteTimeValue(True)
        self.writer.Write()



"""
This file includes the definition and tools to manipulate files
----
Luc Laurent - luc.laurent@lecnam.net -- 2021
"""

import time
from pathlib import Path

from loguru import logger as Logger

from . import various


class fileHandler:
    def __init__(self, filename=None, append=None, right='w', gz=False, bz2=False, safeMode=False):
        """
        create the class
        arguments:
            - filename: name of the file to open
            - append: append to existing file (override 'right')
            - right: specific right when open the file ('r','a','w'...)
            - gz: on the fly compress file with gunzip
            - bz2: on the fly compress file with bunzip2
            - safeMode: avoid overwritten
        """
        self.filename = None
        self.dirname = None
        self.fhandle = None
        self.right = right
        self.append = None
        self.compress = None
        self.startTime = 0
        #
        self.fixRight(append=append, right=right)

        # check arguments
        checkOk = True
        if not filename:
            checkOk = False
            Logger.error('Filename argument missing')
        if not right and not append:
            checkOk = False
            Logger.error('Right(s) not provided')
        # load the filename
        self.getFilename(Path(filename), gz, bz2)
        # open the file
        self.open(safeMode)

    def getFilename(self, filename, gz=None, bz2=None):
        """
        Load the right filename
        """
        self.compress = None
        # check extension for compression
        if filename.suffix == '.gz':
            self.compress = 'gz'
        elif filename.suffix == '.bz2':
            self.compress = 'bz2'
        elif gz:
            filename.with_suffix(filename.suffix + '.gz')
            self.compress = 'gz'
        elif bz2:
            filename.with_suffix(filename.suffix + '.bz2')
            self.compress = 'bz2'
        # extract information about filename
        self.basename = filename.name
        self.dirname = filename.absolute().parent
        self.filename = filename

    def open(self, safeMode=False):
        """
        Open the file w/- or W/o safe mode (avoid overwrite existing file
        """
        # adapt the rights (in case of the file does not exist)
        if self.append and self.filename.exists():
            Logger.warning(f'{self.basename} does not exist! Unable to append')
            self.fixRight(append=False)
        if not safeMode and self.filename.exists() and not self.append and 'w' in self.right:
            Logger.warning(f'{self.basename} already exists! It will be overwritten')
        if safeMode and self.filename.exists() and not self.append and 'w' in self.right:
            Logger.warning(f'{self.basename} already exists! Not overwrite it')
        else:
            #
            Logger.debug(f'Open {self.basename} in {self.dirname} with right {self.right}')
            # open file
            if self.compress == 'gz':
                Logger.debug('Use GZ lib')
                import gzip

                self.fhandle = gzip.open(self.filename, self.right)
            elif self.compress == 'bz2':
                Logger.debug('Use BZ2 lib')
                import bz2

                self.fhandle = bz2.open(self.filename, self.right)
            else:
                self.fhandle = open(self.filename, self.right)
        # store timestamp at opening
        self.startTime = time.perf_counter()
        return self.fhandle

    def close(self):
        """
        Close openned file
        """
        if self.fhandle:
            self.fhandle.close()
            self.fhandle = None
            Logger.info(
                f'Close file {self.basename} with elapsed time {time.perf_counter()-self.startTime:g}s - size {various.convert_size(self.filename.stat().st_size)}'
            )

    def getHandler(self):
        """
        get the file handler
        """
        return self.fhandle

    def write(self, txt):
        """
        write in the file using handle
        """
        return self.fhandle.write(txt)

    def fixRight(self, append=None, right=None):
        """
        Fix issue on right to write file
        """
        if append is not None:
            self.append = append
            if append:
                self.right = 'a'
            else:
                self.right = 'w'
        else:
            self.right = right
            if right[0] == 'w':
                self.append = False
            elif right[0] == 'a':
                self.append = True



import pickle
from pathlib import Path

import numpy

from meshRW import vtk, vtk2

# load current path
CurrentPath = Path(__file__).parent
DataPath = CurrentPath / Path('test_data')
# data file for testing
datafile = DataPath / Path('debug.h5')
# artifacts directory
ArtifactsPath = CurrentPath / Path('artifacts')
ArtifactsPath.mkdir(exist_ok=True)


def test_VTKwriterTemporal():
    # open data
    hf = open(datafile, 'rb')
    data = pickle.load(hf)
    hf.close()
    # extract nodes list
    nodes = data['n']
    # extract elements list and data
    elemsData = data['e']
    # generate data on nodes
    dataNodes = numpy.random.rand(nodes.shape[0], nodes.shape[1])
    # generate data on elements
    dataElem = numpy.random.rand(elemsData['TET4'].shape[0] + elemsData['PRI6'].shape[0], 2)
    # generate steps
    dataElemStep = [numpy.random.rand(elemsData['TET4'].shape[0] + elemsData['PRI6'].shape[0], 3) for i in range(5)]
    # write msh file
    outputfile = ArtifactsPath / Path('build-temp.vtk')
    vtk.vtkWriter(
        filename=outputfile,
        nodes=nodes,
        elements=[
            {
                'connectivity': elemsData[list(elemsData.keys())[0]] - 1,
                'type': list(elemsData.keys())[0],
                'physgrp': [5, 5],
            },
            {
                'connectivity': elemsData[list(elemsData.keys())[1]] - 1,
                'type': list(elemsData.keys())[1],
                'physgrp': [6, 6],
            },
        ],
        fields=[
            {
                'data': dataNodes,
                'type': 'nodal',
                'dim': 3,
                'name': 'nodal3',
            },  # ,'steps':list of steps,'nbsteps':number of steps],
            {'data': dataElem, 'type': 'elemental', 'dim': 2, 'name': 'name_2'},
            {'data': dataElemStep, 'type': 'elemental', 'dim': 3, 'name': 'alongsteps', 'nbsteps': 5},
        ],  # ,'steps':list of steps,'nbsteps':number of steps]
    )
    for i in range(5):
        artifact = outputfile.parent / Path(outputfile.stem + f'.{i:d}' + '.vtk')
        assert artifact.exists()



def test_VTKwriterNoPhysGrp():
    # open data
    hf = open(datafile, 'rb')
    data = pickle.load(hf)
    hf.close()
    # extract nodes list
    nodes = data['n']
    # extract elements list and data
    elemsData = data['e']
    # generate data on nodes
    dataNodes = numpy.random.rand(nodes.shape[0], nodes.shape[1])
    # generate data on elements
    dataElem = numpy.random.rand(elemsData['TET4'].shape[0] + elemsData['PRI6'].shape[0], 2)
    # write msh file
    outputfile = ArtifactsPath / Path('build.vtk')
    vtk.vtkWriter(
        filename=outputfile,
        nodes=nodes,
        elements=[
            {
                'connectivity': elemsData[list(elemsData.keys())[0]] - 1,
                'type': list(elemsData.keys())[0]
            },
            {
                'connectivity': elemsData[list(elemsData.keys())[1]] - 1,
                'type': list(elemsData.keys())[1]
            },
        ]
    )
    assert outputfile.exists()

def test_VTKwriterStatic():
    # open data
    hf = open(datafile, 'rb')
    data = pickle.load(hf)
    hf.close()
    # extract nodes list
    nodes = data['n']
    # extract elements list and data
    elemsData = data['e']
    # generate data on nodes
    dataNodes = numpy.random.rand(nodes.shape[0], nodes.shape[1])
    # generate data on elements
    dataElem = numpy.random.rand(elemsData['TET4'].shape[0] + elemsData['PRI6'].shape[0], 2)
    # write msh file
    outputfile = ArtifactsPath / Path('build.vtk')
    vtk.vtkWriter(
        filename=outputfile,
        nodes=nodes,
        elements=[
            {
                'connectivity': elemsData[list(elemsData.keys())[0]] - 1,
                'type': list(elemsData.keys())[0],
                'physgrp': [5, 5],
            },
            {
                'connectivity': elemsData[list(elemsData.keys())[1]] - 1,
                'type': list(elemsData.keys())[1],
                'physgrp': [6, 6],
            },
        ],
        fields=[
            {
                'data': dataNodes,
                'type': 'nodal',
                'dim': 3,
                'name': 'nodal3',
            },  # ,'steps':list of steps,'nbsteps':number of steps],
            {'data': dataElem, 'type': 'elemental', 'dim': 2, 'name': 'name_2'},
        ],
    )
    assert outputfile.exists()


def test_VTK2writerTemporal():
    # open data
    hf = open(datafile, 'rb')
    data = pickle.load(hf)
    hf.close()
    # extract nodes list
    nodes = data['n']
    # extract elements list and data
    elemsData = data['e']
    # generate data on nodes
    dataNodes = numpy.random.rand(nodes.shape[0], nodes.shape[1])
    # generate data on elements
    dataElem = numpy.random.rand(elemsData['TET4'].shape[0] + elemsData['PRI6'].shape[0], 2)
    # generate steps
    dataElemStep = [numpy.random.rand(elemsData['TET4'].shape[0] + elemsData['PRI6'].shape[0], 3) for i in range(5)]
    # write msh file
    outputfile = ArtifactsPath / Path('build2-temp.vtu')
    extension = outputfile.suffix
    vtk2.vtkWriter(
        filename=outputfile,
        nodes=nodes,
        elements=[
            {
                'connectivity': elemsData[list(elemsData.keys())[0]] - 1,
                'type': list(elemsData.keys())[0],
                'physgrp': [5, 5],
            },
            {
                'connectivity': elemsData[list(elemsData.keys())[1]] - 1,
                'type': list(elemsData.keys())[1],
                'physgrp': [6, 6],
            },
        ],
        fields=[
            {
                'data': dataNodes,
                'type': 'nodal',
                'dim': 3,
                'name': 'nodal3',
            },  # ,'steps':list of steps,'nbsteps':number of steps],
            {'data': dataElem, 'type': 'elemental', 'dim': 2, 'name': 'name_2'},
            {
                'data': dataElemStep,
                'type': 'elemental',
                'dim': 3,
                'name': 'alongsteps',
                'nbsteps': 5,
            },  # ,'steps':list of steps,'nbsteps':number of steps]
        ],
    )
    for i in range(len(dataElemStep)):
        assert outputfile.with_suffix(f'.{i:d}' + extension).exists()
    assert outputfile.with_suffix('.pvd').exists()



def test_VTK2writerNoPhysGrp():
    # open data
    hf = open(datafile, 'rb')
    data = pickle.load(hf)
    hf.close()
    # extract nodes list
    nodes = data['n']
    # extract elements list and data
    elemsData = data['e']
    # generate data on nodes
    dataNodes = numpy.random.rand(nodes.shape[0], nodes.shape[1])
    # generate data on elements
    dataElem = numpy.random.rand(elemsData['TET4'].shape[0] + elemsData['PRI6'].shape[0], 2)
    # write msh file
    outputfile = ArtifactsPath / Path('build.vtk')
    vtk.vtkWriter(
        filename=outputfile,
        nodes=nodes,
        elements=
            {
                'connectivity': elemsData[list(elemsData.keys())[0]] - 1,
                'type': list(elemsData.keys())[0]
            }
        ,
        fields=[
            {
                'data': dataNodes,
                'type': 'nodal',
                'dim': 3,
                'name': 'nodal3',
            },  # ,'steps':list of steps,'nbsteps':number of steps],
            {'data': dataElem, 'type': 'elemental', 'dim': 2, 'name': 'name_2'},
        ]
    )
    assert outputfile.exists()

def test_VTK2writerStatic():
    # open data
    hf = open(datafile, 'rb')
    data = pickle.load(hf)
    hf.close()
    # extract nodes list
    nodes = data['n']
    # extract elements list and data
    elemsData = data['e']
    # generate data on nodes
    dataNodes = numpy.random.rand(nodes.shape[0], nodes.shape[1])
    # generate data on elements
    dataElem = numpy.random.rand(elemsData['TET4'].shape[0] + elemsData['PRI6'].shape[0], 2)
    # write msh file
    outputfile = ArtifactsPath / Path('build2.vtu')
    vtk2.vtkWriter(
        filename=outputfile,
        nodes=nodes,
        elements=[
            {
                'connectivity': elemsData[list(elemsData.keys())[0]] - 1,
                'type': list(elemsData.keys())[0],
                'physgrp': [5, 5],
            },
            {
                'connectivity': elemsData[list(elemsData.keys())[1]] - 1,
                'type': list(elemsData.keys())[1],
                'physgrp': [6, 6],
            },
        ],
        fields=[
            {
                'data': dataNodes,
                'type': 'nodal',
                'dim': 3,
                'name': 'nodal3',
            },  # ,'steps':list of steps,'nbsteps':number of steps],
            {'data': dataElem, 'type': 'elemental', 'dim': 2, 'name': 'name_2'},
        ],
    )

    assert outputfile.exists()



"""
This class is a part of the SILEX library and will write results in legacy VTK file.
----
Luc Laurent - luc.laurent@lecnam.net -- 2021
"""

from pathlib import Path
from typing import Union

import numpy as np
from loguru import logger as Logger

from . import configMESH, dbvtk, fileio, various, writerClass


class vtkWriter(writerClass.writer):
    def __init__(
        self,
        filename: Union[str, Path] = None,
        nodes: Union[list, np.ndarray] = None,
        elements: dict = None,
        fields: Union[list, np.ndarray] = None,
        append: bool = False,
        title: str = None,
        opts: dict = {'version': 'v2'},
    ):
        """
        load class to write file

        syntax:
            vtkWriter(filename,nodes,elems,fields)

        inputs:
            filename : string, name of the gmsh file (with or without msh extension),
                                may contain a directory name
            nodes    : nodes coordinates
            elems : connectivity tables (could contains many kind of elements)
                    list of connectivity dict such as [{connectivity:table1,'type':eltype1,phygrp:gpr1},{connectivity':table2,'type':eltype1,configMESH.DFLT_PHYS_GRP:grp2}...]
                    connectivity: connectivity array
                    'type': type of elements (could be a string or an integer, see getGmshElemType and  gmsh documentation)
                    configMESH.DFLT_PHYS_GRP (optional): physical group (integer or array of integers to declare the physical group of each cell)
            fields (optional)  : list of the fields to write, syntax:
                fields=[{'data':variable_name1,'type':'nodal' or 'elemental' ,'dim':number of values per node/cell,'name':'name 1','steps':list of steps,'nbsteps':number of steps],
                        {'data':variable_name2,'type':'nodal' or 'elemental' ,'dim':number of values per node/cell,'name':'name 2','steps':list of steps,'nbsteps':number of steps],
                        ...
                        ]
            append (optional, default: False) : append field to an existing file
            title (optional, default: None) : title of the file
            opts (optional): dictionary of options (version: VTK's format (v2 (default) or XML)

        """
        Logger.info('Start writing vtk file')
        # adapt inputs
        nodes, elements, fields = writerClass.adaptInputs(nodes, elements, fields)
        # prepare new fields (from physical groups for instance)
        newFields = self.createNewFields(elements)
        if newFields:
            if not fields:
                fields = list()
            fields.extend(newFields)
        # initialization
        super().__init__(filename, nodes, elements, fields, append, title, opts)
        # load specific configuration
        self.db = dbvtk
        # write contents depending on the number of steps
        self.writeContentsSteps(nodes, elements, fields)

    def setOptions(self, options: dict):
        """Default options"""
        self.version = options.get('version', 'v2')

    def writeContentsSteps(self, nodes, elements, fields=None):
        """Write content along steps"""
        # write along steps
        if self.nbSteps > 0:
            for itS in range(self.nbSteps):
                # adapt title
                self.title = self.adaptTitle(txt=f' step num {itS:d}', append=True)
                # adapt the filename
                filename = self.getFilename(suffix='.' + str(itS).zfill(len(str(self.nbSteps))))
                self.customHandler = fileio.fileHandler(filename=filename, append=self.append, safeMode=False)
                # prepare fields (only write all fields on the first step)
                fieldsOk = list()
                fieldsOk = fields
                Logger.info(f'Start writing {self.customHandler.filename}')
                self.writeContents(nodes, elements, fieldsOk, numStep=itS)
                self.customHandler.close()
        else:
            filename = self.getFilename()
            self.customHandler = fileio.fileHandler(filename=filename, append=self.append, safeMode=False)
            Logger.info(f'Start writing {self.customHandler.filename}')
            self.writeContents(nodes, elements, fields)
            self.customHandler.close()

    def writeContents(self, nodes, elements, fields=None, numStep=None):
        """
        Write all contents for one step
        """
        # if we are not appending to an existing file
        if not self.getAppend():
            # write header
            self.writeHeader()
            # write nodes
            self.writeNodes(nodes)
            # write elements
            self.writeElements(elements)
        # write fields if available
        if fields is not None:
            self.writeFields(fields, numStep)

    def getAppend(self):
        """
        Obtain the adapt flag from the handler (automatic adaptation if the file exists)
        """
        self.append = self.customHandler.append
        return self.append

    def logBadExtension(self):
        """ """
        Logger.error('File {}: bad extension (ALLOWED: {})'.format(self.filename, ' '.join(dbvtk.ALLOWED_EXTENSIONS)))

    def writeHeader(self):
        """
        Write header of the VTK file
        """
        if self.version == 'v2':
            headerVTKv2(self.customHandler.fhandle, commentTxt=self.title)
        elif self.version == 'xml':
            self.headerVTKXML(self.customHandler.fhandle)

    @various.timeit('Nodes written')
    def writeNodes(self, nodes):
        """
        Write nodes depending on version
        """
        # count number of nodes
        self.nbNodes = nodes.shape[0]
        if self.version == 'v2':
            WriteNodesV2(self.customHandler.fhandle, nodes)
        elif self.version == 'xml':
            WriteNodesXML(self.customHandler.fhandle, nodes)

    @various.timeit('Elements written')
    def writeElements(self, elems):
        """
        Write elements depending on version
        """
        # convert to list if dict
        if type(elems) is dict:
            elemsRun = [elems]
        else:
            elemsRun = elems
        # count number of elements
        self.nbElems = 0
        for e in elemsRun:
            self.nbElems += e[configMESH.DFLT_MESH].shape[0]

        if self.version == 'v2':
            WriteElemsV2(self.customHandler.fhandle, elems)
        elif self.version == 'xml':
            WriteElemsXML(self.customHandler.fhandle, elems)

    def createNewFields(self, elems):
        """
        Prepare new fields from elems data (for instance physical group)
        """
        # check if physgroup exists
        physGrp = False
        newFields = None
        for itE in elems:
            if configMESH.DFLT_PHYS_GRP in itE.keys():
                physGrp = True
                break
        if physGrp:
            newFields = list()
            data = list()
            for itE in elems:
                nbElems = itE[configMESH.DFLT_MESH].shape[0]
                if configMESH.DFLT_PHYS_GRP in itE.keys():
                    dataPhys = np.array(itE[configMESH.DFLT_PHYS_GRP], dtype=int)
                    if len(dataPhys) == nbElems:
                        data = np.append(data, dataPhys)
                    else:
                        data = np.append(data, dataPhys[0] * np.ones(nbElems))
                else:
                    data = np.append(data, -np.ones(nbElems))
            Logger.debug('Create new field for physical group')
            newFields.extend([{'data': data, 'type': 'elemental_scalar', 'dim': 1, 'name': configMESH.DFLT_PHYS_GRP}])

        return newFields

    @various.timeit('Fields written')
    def writeFields(self, fields, numStep=None):
        """
        Write fields depending on version
        """
        if self.version == 'v2':
            WriteFieldsV2(self.customHandler.fhandle, self.nbNodes, self.nbElems, fields, numStep)
        elif self.version == 'xml':
            WriteFieldsXML(self.customHandler.fhandle, self.nbNodes, self.nbElems, fields, numStep)


# classical function to write contents
# write header in VTK file
def headerVTKv2(fileHandle, commentTxt=''):
    fileHandle.write(f'{dbvtk.DFLT_HEADER_VERSION}\n')
    fileHandle.write(f'{commentTxt}\n')
    fileHandle.write(f'{dbvtk.DFLT_TYPE_ASCII}\n')
    fileHandle.write(f'{dbvtk.DFLT_TYPE_MESH}\n')


def headerVTKXML(fileHandle, commentTxt=''):
    pass


def WriteNodesV2(fileHandle, nodes):
    """Write nodes coordinates for unstructured grid"""
    nbNodes = nodes.shape[0]
    Logger.debug(f'Write {nbNodes} nodes')
    fileHandle.write(f'\n{dbvtk.DFLT_NODES} {nbNodes:d} {dbvtk.DFLT_DOUBLE}\n')
    #
    dimPb = nodes.shape[1]

    # declare format specification
    if dimPb == 2:
        formatSpec = '{:9.4g} {:9.4g}\n'
    elif dimPb == 3:
        formatSpec = '{:9.4g} {:9.4g} {:9.4g}\n'
    # write coordinates
    for i in range(nbNodes):
        fileHandle.write(formatSpec.format(*nodes[i, :]))


def WriteNodesXML(fileHandle, nodes):
    """Write nodes coordinates for unstructured grid"""


def WriteElemsV2(fileHandle, elements):
    """Write elements for unstructured grid"""
    # count data
    nbElems = 0
    nbInt = 0
    for itE in elements:
        nbElems += itE[configMESH.DFLT_MESH].shape[0]
        nbInt += np.prod(itE[configMESH.DFLT_MESH].shape)
        Logger.debug(f'{itE[configMESH.DFLT_MESH].shape[0]} {itE[configMESH.DFLT_FIELD_TYPE]}')

    # initialize size declaration
    fileHandle.write(f'\n{dbvtk.DFLT_ELEMS} {nbElems:d} {nbInt+nbElems:d}\n')
    Logger.debug(f'Start writing {nbElems} {dbvtk.DFLT_ELEMS}')
    # along the element types
    for itE in elements:
        # get the numbering the the element and the number of nodes per element
        nbNodesPerCell = dbvtk.getNumberNodes(itE[configMESH.DFLT_FIELD_TYPE])
        formatSpec = '{:d} '
        formatSpec += ' '.join('{:d}' for _ in range(nbNodesPerCell))
        formatSpec += '\n'
        # write cells
        for e in itE[configMESH.DFLT_MESH]:
            fileHandle.write(formatSpec.format(nbNodesPerCell, *e))

    # declaration of cell types
    fileHandle.write(f'\n{dbvtk.DFLT_ELEMS_TYPE} {nbElems:d}\n')
    Logger.debug(f'Start writing {nbElems} {dbvtk.DFLT_ELEMS_TYPE}')
    # along the element types
    for itE in elements:
        numElemVTK, _ = dbvtk.getVTKElemType(itE[configMESH.DFLT_FIELD_TYPE])
        for _ in range(itE[configMESH.DFLT_MESH].shape[0]):
            fileHandle.write(f'{numElemVTK:d}\n')


def WriteElemsXML(fileHandle, elements):
    """Write elements  for unstructured grid"""


def WriteFieldsV2(fileHandle, nbNodes, nbElems, fields, numStep=None):
    """
        write fields
    input:
        elems: lists of dict of connectivity with elements type (could be reduce to only one dictionary and elements)
                [{'connectivity':table1,'type':eltype1,physgrp:gpr1},{'connectivity':table2,'type':eltype1,configMESH.DFLT_PHYS_GRP:grp2}...]
                or
                {'connectivity':table1,'type':eltype1,'physgrp':gpr1}

                'connectivity': connectivity array
                'type': type of elements (could be a string or an integer, see getGmshElemType and  gmsh documentation)
                'physgrp' (optional): physical group (integer or array of integers to declare the physical group of each cell)
        fields=[{'data':variable_name1,'type':'nodal' or 'elemental' ,'dim':number of values per node,'name':'name 1','steps':list of steps,'nbsteps':number of steps],
                    {'data':variable_name2,'type':'nodal' or 'elemental' ,'dim':number of values per node,'name':'name 2','steps':list of steps,'nbsteps':number of steps],
                    ...
                    ]

                'data': array of the data or list of dictionary
                'type': ('nodal' or 'elemental') data given at nodes or cells
                'dim': number of data per nodes/cells
                'name': name of the data
                'steps' (optional): list of steps used to declare fields
                'nbsteps' (optional): number of steps used to declare fields
                if no 'steps' or 'nbsteps' are declared the field is assumed to be not defined along steps
                #
                'data' could be defined as
                     - list of a arrays with all nodal or elemental values along steps
                     - a dictionary {'array':ar,'connectivityId':int} in the case of elemental
                        'connectivityId': the data are given associated to a certain list of cells (other is defined as 0)

    """
    # analyze fields data
    iXNodalField = list()
    iXElementalField = list()
    iXNodalScalar = list()
    iXElementalScalar = list()
    for i, f in enumerate(fields):
        if f[configMESH.DFLT_FIELD_TYPE] == configMESH.DFLT_FIELD_TYPE_NODAL:
            iXNodalField.append(i)
        elif f[configMESH.DFLT_FIELD_TYPE] == configMESH.DFLT_FIELD_TYPE_ELEMENT:
            iXElementalField.append(i)
        elif f[configMESH.DFLT_FIELD_TYPE] == configMESH.DFLT_FIELD_TYPE_NODAL_SCALAR:
            iXNodalScalar.append(i)
        elif f[configMESH.DFLT_FIELD_TYPE] == configMESH.DFLT_FIELD_TYPE_ELEMENT_SCALAR:
            iXElementalScalar.append(i)

    # write CELL_DATA
    if len(iXElementalField) + len(iXElementalScalar) > 0:
        Logger.debug(f'Start writing {nbElems} {dbvtk.DFLT_ELEMS_DATA}')
        fileHandle.write(f'\n{dbvtk.DFLT_ELEMS_DATA} {nbElems:d}\n')

        # write scalars
        if len(iXElementalScalar) > 0:
            for iX in iXElementalScalar:
                # get array of data
                data = getData(fields[iX], numStep)
                writeScalarsDataV2(fileHandle, data, fields[iX]['name'])
        # write fields
        if len(iXElementalField) > 0:
            Logger.debug(f'Start writing {len(iXElementalField)} {dbvtk.DFLT_FIELD}')
            fileHandle.write('{} {} {:d}\n'.format(dbvtk.DFLT_FIELD, 'cellField', len(iXElementalField)))
            for iX in iXElementalField:
                # get array of data
                data = getData(fields[iX], numStep)
                writeFieldsDataV2(fileHandle, data, fields[iX]['name'])

    # write POINT_DATA
    if len(iXNodalField) + len(iXNodalScalar) > 0:
        Logger.debug(f'Start writing {nbNodes} {dbvtk.DFLT_NODES_DATA}')
        fileHandle.write(f'\n{dbvtk.DFLT_NODES_DATA} {nbNodes:d}\n')
        # write scalars
        if len(iXNodalScalar) > 0:
            for iX in iXNodalScalar:
                # get array of data
                data = getData(fields[iX], numStep)
                writeScalarsDataV2(fileHandle, data, fields[iX]['name'])
        # write fields
        if len(iXNodalField) > 0:
            Logger.debug(f'Start writing {len(iXNodalField)} {dbvtk.DFLT_FIELD}')
            fileHandle.write('{} {} {:d}\n'.format(dbvtk.DFLT_FIELD, 'pointField', len(iXNodalField)))
            for iX in iXNodalField:
                # get array of data
                data = getData(fields[iX], numStep)
                writeFieldsDataV2(fileHandle, data, fields[iX]['name'])


def getData(data, num):
    """
    get data for the right step
    """
    # create array of data
    dataOut = None
    if configMESH.DFLT_FIELD_STEPS in data.keys():
        if len(data[configMESH.DFLT_FIELD_STEPS]) > 1:
            dataOut = data[configMESH.DFLT_FIELD_DATA][num]
    elif configMESH.DFLT_FIELD_NBSTEPS in data.keys():
        if data[configMESH.DFLT_FIELD_NBSTEPS] > 0:
            dataOut = data[configMESH.DFLT_FIELD_DATA][num]
    else:
        dataOut = data[configMESH.DFLT_FIELD_DATA]
    return dataOut


def writeScalarsDataV2(fileHandle, data, name):
    """
    write data using SCALARS
    """
    if len(data.shape) > 1:
        nbComp = data.shape[1]
    else:
        nbComp = 1
    # dataType
    dataType = 'double'
    formatSpec = ' '.join('{:9.4f}' for _ in range(nbComp)) + '\n'
    if issubclass(data.dtype.type, np.integer):
        dataType = 'int'
        formatSpec = ' '.join('{:d}' for _ in range(nbComp)) + '\n'
    elif issubclass(data.dtype.type, np.floating):
        dataType = 'double'
        formatSpec = ' '.join('{:9.4f}' for _ in range(nbComp)) + '\n'
    Logger.debug(f'Start writing {dbvtk.DFLT_SCALARS} {name}')
    fileHandle.write(f'{dbvtk.DFLT_SCALARS} {name} {dataType} {nbComp:d}\n')
    fileHandle.write(f'{dbvtk.DFLT_TABLE} {dbvtk.DFLT_TABLE_DEFAULT}\n')
    for d in data:
        fileHandle.write(formatSpec.format(d))


def writeFieldsDataV2(fileHandle, data, name):
    """
    write data using FIELD
    """
    nbComp = data.shape[1]
    # dataType
    dataType = 'double'
    formatSpec = ' '.join('{:9.4f}' for _ in range(nbComp)) + '\n'
    if issubclass(data.dtype.type, np.integer):
        dataType = 'int'
        formatSpec = ' '.join('{:d}' for _ in range(nbComp)) + '\n'
    elif issubclass(data.dtype.type, np.floating):
        dataType = 'double'
        formatSpec = ' '.join('{:9.4f}' for _ in range(nbComp)) + '\n'
    # start writing
    Logger.debug(f'Start writing {dbvtk.DFLT_FIELD} {name}')
    fileHandle.write(f'{name} {nbComp:d} {data.shape[0]:d} {dataType}\n')
    for d in data:
        fileHandle.write(formatSpec.format(*d))


def WriteFieldsXML(fileHandle, nbNodes, nbElems, fields, numStep=None):
    """Write elements"""



""" "
This file includes the definition and tools to manipulate MSH format
Documentation available here: https://gmsh.info/doc/texinfo/gmsh.html#MSH-file-format
----
Luc Laurent - luc.laurent@lecnam.net -- 2021

"""

import vtk
from loguru import logger as Logger


def loadElementDict():
    """
    dictionary from element (string) to VTK number
    """
    elementDict = {
        # 2-nodes line
        'LIN2': {'code': 3, 'nodes': 2, 'dim': 1, 'vtkobj': vtk.vtkLine()},
        # 3-nodes second order line
        'LIN3': {'code': 21, 'nodes': 3, 'dim': 1, 'vtkobj': vtk.vtkQuadraticEdge()},
        # 4-nodes third order line
        'LIN4': None,
        # 3-nodes triangle
        'TRI3': {'code': 5, 'nodes': 3, 'dim': 2, 'vtkobj': vtk.vtkTriangle()},
        # 6-nodes second order triangle (3 vertices, 3 on edges)
        'TRI6': {'code': 22, 'nodes': 6, 'dim': 2, 'vtkobj': vtk.vtkQuadraticTriangle()},
        # 9-nodes cubic order triangle (3 vertices, 3 on edges and 3 inside)
        'TRI9': None,
        # 10-nodes higher order triangle (3 vertices, 6 on edges and 1 inside)
        'TRI10': None,
        # 12-nodes higher order triangle (3 vertices and 9 on edges)
        'TRI12': None,
        # 15-nodes higher order triangle (3 vertices, 9 on edges and 3 inside)
        'TRI15': None,
        # 4-nodes quadrangle
        'QUA4': {'code': 9, 'nodes': 4, 'dim': 2, 'vtkobj': vtk.vtkQuad()},
        # 8-nodes second order quadrangle (4 vertices and 4 on edges)
        'QUA8': {'code': 23, 'nodes': 8, 'dim': 2, 'vtkobj': vtk.vtkQuadraticQuad()},
        # 9-nodes higher order quadrangle (4 vertices, 4 on edges and 1 inside)
        'QUA9': None,
        # 4-nodes tetrahedron
        'TET4': {'code': 10, 'nodes': 4, 'dim': 3, 'vtkobj': vtk.vtkTetra()},
        # 10-nodes second order tetrahedron (4 vertices and 6 on edges)
        'TET10': {'code': 24, 'nodes': 10, 'dim': 3, 'vtkobj': vtk.vtkQuadraticTetra()},
        # 8-nodes hexahedron
        'HEX8': {'code': 12, 'nodes': 8, 'dim': 3, 'vtkobj': vtk.vtkHexahedron()},
        # 20-nodes second order hexahedron (8 vertices and 12 on edges)
        'HEX20': {'code': 25, 'nodes': 20, 'dim': 3, 'vtkobj': vtk.vtkQuadraticHexahedron()},
        # 27-nodes higher order hexahedron (8 vertices,
        # 12 on edges, 6 on faces and 1 inside)
        'HEX27': None,
        # 6-nodes prism
        'PRI6': {'code': 13, 'nodes': 6, 'dim': 3, 'vtkobj': vtk.vtkWedge()},
        # 15-nodes second order prism (6 vertices and 9 on edges)
        'PRI15': None,
        # 18-nodes higher order prism (6 vertices, 9 on edges and 3 on faces)
        'PRI18': None,
        # 5-node pyramid
        'PYR5': {'code': 14, 'nodes': 5, 'dim': 3, 'vtkobj': vtk.vtkPyramid()},
        # 13-nodes second order pyramid (5 edges and 8 on edges)
        'PYR13': None,
        # 14-nodes higher order pyramid (5 edges, 8 on edges and 1 inside)
        'PYR14': None,
        # 1-node point
        'NOD1': {'code': 1, 'nodes': 1, 'dim': 0, 'vtkobj': vtk.vtkVertex()},
        #
        # many nodes
        'NODN': {'code': 2, 'nodes': -1, 'dim': 0, 'vtkobj': vtk.vtkPolyVertex()},
        # many lines (poly-lines)
        'LINEN': {'code': 4, 'nodes': -1, 'dim': 1, 'vtkobj': vtk.vtkPolyLine()},
        # many stripped triangles
        'TRIN': {'code': 6, 'nodes': -1, 'dim': 2, 'vtkobj': vtk.vtkTriangleStrip()},
        # polygons
        'POLY': {'code': 7, 'nodes': -1, 'dim': 2, 'vtkobj': vtk.vtkPolygon()},
        # pixel
        'PIXEL': {'code': 8, 'nodes': -1, 'dim': 2, 'vtkobj': vtk.vtkPixel()},
        # voxel
        'VOXEL': {'code': 11, 'nodes': -1, 'dim': 3, 'vtkobj': vtk.vtkVoxel()},
    }
    return elementDict


def getVTKtoElem():
    """
    dictionary from VTK element to name of element (string)
    """
    VTKtoElem = {
        'VTK_VERTEX': 'NOD1',
        'VTK_LINE': 'LIN2',
        'VTK_TRIANGLE': 'TRI3',
        'VTK_QUAD': 'QUAD4',
        'VTK_TETRA': 'TET4',
        'VTK_HEXAHEDRON': 'HEX8',
        'VTK_WEDGE': 'PRI6',
        'VTK_PYRAMID': 'PYR5',
        'VTK_QUADRATIC_EDGE': 'LIN3',
        'VTK_QUADRATIC_TRIANGLE': 'TRI6',
        'VTK_QUADRATIC_QUAD': 'QUA8',
        'VTK_QUADRATIC_TETRA': 'TET10',
        'VTK_QUADRATIC_HEXAHEDRON': 'HEX20',
        #
        'VTK_POLY_VERTEX': 'NODN',
        'VTK_POLY_LINE': 'LINEN',
        'VTK_TRIANGLE_STRIP': 'TRIN',
        'VTK_POLYGON': 'POLY',
        'VTK_PIXEL': 'PIXEL',
        'VTK_VOXEL': 'VOXEL',
    }
    return VTKtoElem


def getVTKObj(txtEltype):
    """
    Get the vtk class from libvtk from text declaration
    syntax:
        getVTKEObjType(txtEltype)

    input:
        txtEltype: element declared using VTK string (if number is used the function wil return it)
    output:
        vtk class for the requested element
        number of nodes on element
    """

    VTKtoElem = getVTKtoElem()
    elementDict = loadElementDict()

    # depending on the type of txtEltype
    numPerElement = -1
    if txtEltype.upper() in VTKtoElem.keys():
        txtEltype = VTKtoElem[txtEltype]
    vtkobj = elementDict[txtEltype.upper()].get('vtkobj', None)
    numPerElement = getNumberNodes(txtEltype.upper())
    if not vtkobj:
        Logger.error(f'Element type {txtEltype} not implemented')
    return vtkobj, numPerElement


def getVTKElemType(txtEltype):
    """
    Get the element type defined as in vtk (refer to VTK documentation for numbering) from text declaration
    syntax:
        getVTKElemType(txtEltype)

    input:
        txtEltype: element declared using VTK string (if number is used the function wil return it)
    output:
        element type defined as VTK number
    """

    VTKtoElem = getVTKtoElem()
    elementDict = loadElementDict()

    # depending on the type of txtEltype
    numPerElement = -1
    if isinstance(txtEltype, int):
        elementNum = txtEltype
    else:
        if txtEltype.upper() in VTKtoElem.keys():
            txtEltype = VTKtoElem[txtEltype]
        elementNum = elementDict[txtEltype.upper()].get('code', None)
        numPerElement = getNumberNodes(txtEltype.upper())
    if not elementNum:
        Logger.error(f'Element type {txtEltype} not implemented')
    return elementNum, numPerElement


def getElemTypeFromVTK(elementNum):
    """
    Get the element type from id (integer) defined in gmsh (refer to gmsh documentation for numbering)
    syntax:
        getElemTypeFromVTK(elementNum)

    input:
        elementNum: integer used in gmsh to declare element
    output:
        global name of the element
    """
    # load the dictionary
    elementDict = loadElementDict()
    globalName = None
    # get the name of the element using the integer iD along the dictionary
    for k, v in elementDict.items():
        if v:
            if v.get('code', None) == elementNum:
                globalName = k
                break
    # if the name of the element if not available show error
    if globalName is None:
        Logger.error(f'Element type not found with id {elementNum}')
    return globalName


def getNumberNodes(txtElemtype):
    """
    Get the number of nodes for a specific element type type (declare as string)
        syntax:
            getNumberNodes(txtElemtype)

       input:
            txtElemtype: element declared using string (if number is used the function wil return it)
        output:
            number of nodes for txtEltype
    """

    elementDict = loadElementDict()
    # check if the type of element exists
    if txtElemtype in elementDict.keys():
        # get the number of nodes for the type of element
        if elementDict[txtElemtype]:
            nbNodes = elementDict[txtElemtype].get('nodes', None)
    else:
        # show error message if the type of element does not exist
        Logger.error(f'Element type {txtElemtype} not defined')
    return nbNodes


def getNumberNodesFromNum(elementNum):
    """
    Get the number of nodfs for a specific element type type (declare as string)
        syntax:
            getNumberNodesFromNum(elementNum)

        input:
            elementNum: integer used in gmsh to declare element
        output:
            number of nodes for txtEltype
    """

    return getNumberNodes(getElemTypeFromVTK(elementNum))


# DEFAULT VALUES
ALLOWED_EXTENSIONS = [
    '.vtk',
    '.vtk.bz2',
    '.vtk.gz',
    '.vtu',
    '.vtu.bz2',
    '.vtu.gz',
]


# Keywords MSH
DFLT_HEADER_VERSION = '# vtk DataFile Version 2.0'
DFLT_TYPE_ASCII = 'ASCII'
DFLT_FILE_VERSION = '2.2 0 8'
DFLT_TYPE_MESH = 'DATASET UNSTRUCTURED_GRID'
DFLT_NODES = 'POINTS'
DFLT_NODES_DATA = 'POINT_DATA'
DFLT_DOUBLE = 'double'
DFLT_FLOAT = 'float'
DFLT_INT = 'int'
DFLT_ELEMS = 'CELLS'
DFLT_ELEMS_TYPE = 'CELL_TYPES'
DFLT_ELEMS_DATA = 'CELL_DATA'
DFLT_FIELD = 'FIELD'
DFLT_SCALARS = 'SCALARS'
DFLT_TABLE = 'LOOKUP_TABLE'
DFLT_TABLE_DEFAULT = 'default'



"""
This file contains the default
declaration of the keywords
used to declare/manipulate a mesh in Python
----
Luc Laurent - luc.laurent@lecnam.net -- 2021
"""

DFLT_MESH = 'connectivity'
DFLT_TYPE_ELEM = 'type'
DFLT_PHYS_GRP = 'physgrp'
DFLT_FIELD_DATA = 'data'
DFLT_FIELD_TYPE = 'type'
DFLT_FIELD_TYPE_NODAL = 'nodal'
DFLT_FIELD_TYPE_ELEMENT = 'elemental'
DFLT_FIELD_TYPE_NODAL_SCALAR = 'nodal_scalar'
DFLT_FIELD_TYPE_ELEMENT_SCALAR = 'elemental_scalar'
DFLT_FIELD_DIM = 'dim'
DFLT_FIELD_NAME = 'name'
DFLT_FIELD_STEPS = 'steps'
DFLT_FIELD_NBSTEPS = 'nbsteps'
DFLT_FIELD_NUMENTITIES = 'numentities'
